# Copyright The Caikit Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


"""Config methods for the `caikit` library. Mainly interacts with `config.yml`.
"""

# Standard
from typing import Any, Dict, Optional, Union
import os

# First Party
import aconfig
import alog

log = alog.use_channel("CONFIG")

BASE_CONFIG_PATH = os.path.realpath(
    os.path.join(os.path.dirname(__file__), "config.yml")
)

# The core config object that is continually merged into
_CONFIG: aconfig.Config = aconfig.Config({})
# An immutable view into the core config object, to be passed to callers
_IMMUTABLE_CONFIG: aconfig.ImmutableConfig = aconfig.ImmutableConfig({})
# Little helper type for signatures
_CONFIG_TYPE = Union[dict, aconfig.Config]


def get_config() -> aconfig.Config:
    """Get the caikit configuration"""
    return _IMMUTABLE_CONFIG


def configure(
    config_yml_path: Optional[str] = None, config_dict: Optional[Dict[str, Any]] = None
):
    """Configure caikit for your usage!
    Merges into the internal aconfig.Config object with overrides from multiple sources.

    Sources, last takes precedence:
        1. The existing configuration from calls to `caikit.configure()`
        2. The config from `config_yml_path`
        3. The config from `config_dict`
        4. The config files specified in the `config_files` configuration
            (NB: This may be set by the `CONFIG_FILES` environment variable)
        5. Environment variables, in ALL_CAPS_SNAKE_FORMAT

    Args:
        config_yml_path (Optional[str]): The path to the base configuration yaml
            with overrides for your usage.
        config_dict (Optional[Dict]): Config overrides in dictionary form

    Returns: None: This only sets the config object that is returned by
        `caikit.get_config()`
    """
    if not config_yml_path and not config_dict:
        log.error("<CFG43273054E>", "No config_file or config_dict provided")
        raise ValueError("No config_file or config_dict provided")

    cfg = aconfig.Config(_CONFIG)
    if config_yml_path:
        new_config = aconfig.Config.from_yaml(config_yml_path)
    else:
        new_config = aconfig.Config(config_dict)

    cfg = merge_configs(cfg, new_config, _get_merge_strategy(new_config))

    cfg = _merge_extra_files(cfg)
    _update_global_config(cfg)


def _update_global_config(cfg: aconfig.Config):
    """Replaces the caikit config and creates a new immutable view of it to be shared via
    get_config().
    """
    # pylint: disable=global-statement
    global _IMMUTABLE_CONFIG
    # pylint: disable=global-statement
    global _CONFIG
    _CONFIG = cfg
    # Set override_env_vars=False because we want the immutable config to be an exact copy
    _IMMUTABLE_CONFIG = aconfig.ImmutableConfig(_CONFIG, override_env_vars=False)


def _merge_extra_files(config: aconfig.Config) -> aconfig.Config:
    """Looks at the `config_files` configuration item and merges those files into the config,
    left to right"""
    if config.config_files:
        extra_config_files = [
            s.strip()
            for s in str(config.config_files or os.environ.get("CONFIG_FILES")).split(
                ","
            )
        ]
        for file in extra_config_files:
            log.info(
                {
                    "log_code": "<RUN17612094I>",
                    "message": "Loading config file '%s'" % file,
                }
            )
            new_overrides = aconfig.Config.from_yaml(file, override_env_vars=True)
            config = merge_configs(
                config, new_overrides, _get_merge_strategy(new_overrides)
            )
    return config


def merge_configs(
    base: Optional[_CONFIG_TYPE],
    overrides: Optional[_CONFIG_TYPE],
    merge_strategy: str = "merge",
) -> _CONFIG_TYPE:
    """Helper to perform a deep merge of the overrides into the base. The merge
    is done in place, but the resulting dict is also returned for convenience.
    The merge logic is quite simple: If both the base and overrides have a key
    and the type of the key for both is a dict, recursively merge, otherwise
    set the base value to the override value.
    Args:
        base (Optional[dict]): The base config that will be updated with the
            overrides
        overrides (Optional[dict]): The override config
        merge_strategy (str): The merging strategy, either `merge` or `override`
            `override` will replace values in base with those from overrides
            `merge` will deep-merge dictionaries and prepend-merge lists
    Returns:
        merged: dict
            The merged results of overrides merged onto base
    """
    # Handle none args
    if base is None:
        return overrides or {}
    if overrides is None:
        return base or {}

    if merge_strategy == "override":
        base.update(overrides)
        return base

    # Do the deep merge
    for key, value in overrides.items():
        if (
            key not in base
            or not isinstance(base[key], (dict, list))
            or not isinstance(value, (dict, list))
        ):
            base[key] = value
        elif isinstance(value, list):
            # merge lists by prepending new one
            base[key] = value + base[key]
        else:
            base[key] = merge_configs(base[key], value, merge_strategy)

    return base


def _get_merge_strategy(cfg: _CONFIG_TYPE) -> str:
    return cfg.get("merge_strategy", "merge")


# Run initial configuration with the base config
configure(BASE_CONFIG_PATH)



# Copyright The Caikit Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Base class to initialize backend
"""

# Standard
from contextlib import contextmanager
from threading import Lock
from typing import Optional
import abc

# First Party
import aconfig


class BackendBase(abc.ABC):
    """Interface for creating configuration setup for backends"""

    def __init__(self, config: Optional[aconfig.Config] = None) -> None:
        self.config = config or {}
        self._started = False
        self._start_lock = Lock()

    @property
    @classmethod
    @abc.abstractmethod
    def backend_type(cls):
        """Property storing type of the backend"""

    @property
    def is_started(self):
        return self._started

    @abc.abstractmethod
    def register_config(self, config):
        """Function to allow dynamic merging of configs.
        This can be useful, if there are explicit configurations
        particular implementations (modules) need to register before the starting the backend.
        """
        # NOTE: This function should be implemented in such a way that it can be called multiple
        # times

    @abc.abstractmethod
    def start(self):
        """Function to start a distributed backend. This function
        should set self._started variable to True"""

    @abc.abstractmethod
    def stop(self):
        """Function to stop a distributed backend. This function
        should set self._started variable to False"""

    @contextmanager
    def start_lock(self):
        with self._start_lock:
            yield



# Copyright The Caikit Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# Standard
from concurrent.futures import Future
from contextlib import contextmanager
from unittest import mock
import tempfile

# Third Party
import grpc
import pytest

# Local
from caikit.config import get_config
from caikit.core import ModuleConfig
from caikit.core.module_backends import backend_types
from caikit.core.modules import base, module
from caikit.runtime.model_management.batcher import Batcher
from caikit.runtime.model_management.model_loader import ModelLoader
from caikit.runtime.types.caikit_runtime_exception import CaikitRuntimeException
from sample_lib.data_model import SampleInputType, SampleOutputType
from sample_lib.modules.sample_task import SampleModule
from tests.conftest import TempFailWrapper, random_test_id, temp_config
from tests.core.helpers import MockBackend
from tests.fixtures import Fixtures
import caikit.core

## Helpers #####################################################################


@contextmanager
def temp_model_loader():
    """Temporarily reset the ModelLoader singleton"""
    real_singleton = ModelLoader.get_instance()
    ModelLoader._ModelLoader__instance = None
    yield ModelLoader.get_instance()
    ModelLoader._ModelLoader__instance = real_singleton


@pytest.fixture
def model_loader():
    return ModelLoader.get_instance()


def make_model_future(model_instance):
    fake_future = Future()
    fake_future.result = lambda *_, **__: model_instance
    return fake_future


## Tests #######################################################################


def test_load_model_ok_response(model_loader):
    """Test that we can load up a valid model folder"""
    model_id = "happy_load_test"
    loaded_model = model_loader.load_model(
        model_id=model_id,
        local_model_path=Fixtures.get_good_model_path(),
        model_type=Fixtures.get_good_model_type(),
    )
    assert loaded_model.model() is not None
    assert isinstance(loaded_model.model(), base.ModuleBase)
    assert model_id == loaded_model.id()
    assert Fixtures.get_good_model_type() == loaded_model.type()
    assert Fixtures.get_good_model_path() == loaded_model.path()

    # Models are not sized by the loader
    assert loaded_model.size() == 0


def test_load_model_archive(model_loader):
    """Test that we can load up a valid model archive"""
    model_id = "happy_load_test"
    loaded_model = model_loader.load_model(
        model_id=model_id,
        local_model_path=Fixtures.get_good_model_archive_path(),
        model_type=Fixtures.get_good_model_type(),
    )
    assert loaded_model.model() is not None
    assert isinstance(loaded_model.model(), base.ModuleBase)


def test_load_model_error_not_found_response(model_loader):
    """Test load model's model does not exist error response"""
    model_id = random_test_id()
    with pytest.raises(CaikitRuntimeException) as context:
        model_loader.load_model(
            model_id=model_id,
            local_model_path="test/this/does/not/exist.zip",
            model_type="categories_esa",
        ).wait()
    assert context.value.status_code == grpc.StatusCode.NOT_FOUND
    assert model_id in context.value.message


def test_load_invalid_model_error_response(model_loader):
    """Test load invalid model error response"""
    model_id = random_test_id()
    with pytest.raises(CaikitRuntimeException) as context:
        model_loader.load_model(
            model_id=model_id,
            local_model_path=Fixtures.get_bad_model_archive_path(),
            model_type="not_real",
        ).wait()
    assert context.value.status_code == grpc.StatusCode.INTERNAL
    assert model_id in context.value.message


def test_it_can_load_more_than_one_model(model_loader):
    """Make sure we can load multiple models without side effects"""
    # TODO: change test to load multiple models

    model_id = "concurrent_load_test"
    model_1 = model_loader.load_model(
        model_id,
        Fixtures.get_good_model_path(),
        model_type=Fixtures.get_good_model_type(),
    )
    model_id = "concurrent_load_test_2"
    model_2 = model_loader.load_model(
        model_id,
        Fixtures.get_good_model_path(),
        model_type=Fixtures.get_good_model_type(),
    )

    assert model_1 is not None
    assert model_2 is not None
    # Different refs
    assert model_1 != model_2


def test_nonzip_extract_fails(model_loader):
    """Check that we raise an error if we throw in an archive that isn't really an archive"""
    model_id = "will_not_be_created"

    with pytest.raises(CaikitRuntimeException) as context:
        model_loader.load_model(
            model_id,
            Fixtures.get_invalid_model_archive_path(),
            Fixtures.get_good_model_type(),
        ).wait()
    # This ends up returning a FileNotFoundError from caikit core.
    # maybe not the best? But it does include an error message at least
    assert (
        context.value.status_code == grpc.StatusCode.NOT_FOUND
    ), "Non-zip file did not raise an error"
    assert Fixtures.get_invalid_model_archive_path() in context.value.message
    assert "config.yml" in context.value.message


def test_no_double_instantiation():
    """Make sure trying to re-instantiate this singleton raises"""
    with pytest.raises(Exception):
        ModelLoader()


def test_with_batching(model_loader):
    """Make sure that loading with batching configuration correctly wraps a
    Batcher around the model.
    """
    model = model_loader.load_model(
        "load_with_batch",
        Fixtures.get_good_model_path(),
        model_type="fake_batch_module",
    ).model()
    assert isinstance(model, Batcher)
    assert model._batch_size == get_config().runtime.batching.fake_batch_module.size

    # Make sure another model loads without batching
    model = model_loader.load_model(
        "load_without_batch",
        Fixtures.get_good_model_path(),
        model_type=Fixtures.get_good_model_type(),
    ).model()
    assert not isinstance(model, Batcher)


def test_with_batching_by_default(model_loader):
    """Make sure that a model type without specific batching enabled will
    load with a batcher if default is enabled
    """
    with temp_config({"runtime": {"batching": {"default": {"size": 10}}}}) as cfg:
        model = model_loader.load_model(
            "load_with_batch_default",
            Fixtures.get_good_model_path(),
            model_type=Fixtures.get_good_model_type(),
        ).model()
        assert isinstance(model, Batcher)
        assert model._batch_size == cfg.runtime.batching.default.size


def test_with_batching_collect_delay(model_loader):
    """Make sure that a non-zero collect_delay_s is read correctly"""
    model_type = Fixtures.get_good_model_type()
    with temp_config(
        {
            "runtime": {
                "batching": {
                    model_type: {
                        "size": 10,
                        "collect_delay_s": 0.01,
                    },
                }
            }
        }
    ) as cfg:
        model = model_loader.load_model(
            "load_with_batch_default",
            Fixtures.get_good_model_path(),
            model_type=model_type,
        ).model()
        assert isinstance(model, Batcher)
        assert model._batch_size == getattr(cfg.runtime.batching, model_type).size
        assert (
            model._batch_collect_delay_s
            == getattr(cfg.runtime.batching, model_type).collect_delay_s
        )


def test_load_distributed_impl(reset_globals):
    """Make sure that when configured, an alternate distributed
    implementation of a module can be loaded
    """

    @module(
        base_module=SampleModule,
        backend_type=backend_types.MOCK,
        backend_config_override={"bar1": 1},
    )
    class DistributedGadget(caikit.core.ModuleBase):
        """An alternate implementation of a Gadget"""

        SUPPORTED_LOAD_BACKENDS = [
            MockBackend.backend_type,
            backend_types.LOCAL,
        ]

        def __init__(self, bar):
            self.bar = bar

        def run(self, sample_input: SampleInputType) -> SampleOutputType:
            return SampleOutputType(greeting=f"hello distributed {sample_input.name}")

        @classmethod
        def load(cls, model_load_path, **kwargs) -> "DistributedGadget":
            # NOTE: kwargs needed here for load_backend
            config = ModuleConfig.load(model_load_path)
            return cls(bar=config.bar)

    with tempfile.TemporaryDirectory() as model_path:
        # Create and save the model directly with the local impl
        SampleModule().save(model_path)

        model_type = "gadget"

        with temp_model_loader() as model_loader:
            # Load the distributed version
            model = model_loader.load_model(
                random_test_id(),
                model_path,
                model_type=model_type,
            ).model()
            assert isinstance(model, DistributedGadget)


def test_load_model_without_waiting_success(model_loader):
    """Make sure that loading a model can defer the model to a future and access
    the loaded model when complete
    """
    model_id = random_test_id()
    loaded_model = model_loader.load_model(
        model_id=model_id,
        local_model_path=Fixtures.get_good_model_path(),
        model_type=Fixtures.get_good_model_type(),
    )
    # Block to get the model
    assert loaded_model.model() is not None
    assert isinstance(loaded_model.model(), base.ModuleBase)
    assert model_id == loaded_model.id()
    assert Fixtures.get_good_model_type() == loaded_model.type()
    assert Fixtures.get_good_model_path() == loaded_model.path()

    # Models are not sized by the loader
    assert loaded_model.size() == 0


def test_load_model_without_waiting_deferred_error(model_loader):
    """Make sure that loading a model can defer the model to a future and raise
    when the future is used if the loading failed
    """
    model_id = random_test_id()
    loaded_model = model_loader.load_model(
        model_id=model_id,
        local_model_path=Fixtures.get_bad_model_archive_path(),
        model_type="not_real",
    )
    with pytest.raises(CaikitRuntimeException) as context:
        loaded_model.model()
    assert context.value.status_code == grpc.StatusCode.INTERNAL
    assert model_id in context.value.message


def test_load_model_succeed_after_retry(model_loader):
    """Make sure that a model which fails to load on the first try can load
    successfully on a retry.
    """
    failures = 2
    fail_wrapper = TempFailWrapper(
        model_loader._load_module,
        num_failures=failures,
        exc=CaikitRuntimeException(grpc.StatusCode.INTERNAL, "Yikes!"),
    )
    with mock.patch.object(model_loader, "_load_module", fail_wrapper):
        model_id = random_test_id()
        loaded_model = model_loader.load_model(
            model_id=model_id,
            local_model_path=Fixtures.get_good_model_path(),
            model_type=Fixtures.get_good_model_type(),
            retries=failures + 1,
        )
        assert loaded_model.model() is not None
        assert isinstance(loaded_model.model(), base.ModuleBase)
        assert loaded_model._retries == 1


def test_load_model_fail_callback_once(model_loader):
    """Make sure that a model which fails all of its retries will only call the
    fail callback once
    """
    failures = 3
    fail_wrapper = TempFailWrapper(
        model_loader._load_module,
        num_failures=failures,
        exc=CaikitRuntimeException(grpc.StatusCode.INTERNAL, "Yikes!"),
    )
    with mock.patch.object(model_loader, "_load_module", fail_wrapper):
        model_id = random_test_id()
        fail_cb = mock.MagicMock()
        loaded_model = model_loader.load_model(
            model_id=model_id,
            local_model_path=Fixtures.get_good_model_path(),
            model_type=Fixtures.get_good_model_type(),
            fail_callback=fail_cb,
            retries=failures - 1,
        )
        with pytest.raises(CaikitRuntimeException):
            loaded_model.wait()
        fail_cb.assert_called_once()



# Copyright The Caikit Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Unit tests for ImportableFactory"""

# Third Party
import pytest

# Local
from caikit.core.model_management.factories import ImportableFactory
from caikit.core.model_management.local_model_finder import LocalModelFinder


def test_importable_factory_auto_register():
    """Test that a constructible gets auto-registered with a factory if given
    the import_class key
    """
    fact = ImportableFactory("TestFact")
    assert not fact._registered_types
    inst = fact.construct(
        {
            ImportableFactory.IMPORT_CLASS_KEY: "{}.{}".format(
                LocalModelFinder.__module__,
                LocalModelFinder.__name__,
            ),
            ImportableFactory.TYPE_KEY: LocalModelFinder.name,
        }
    )
    assert inst
    assert fact._registered_types


@pytest.mark.parametrize(
    "params",
    [
        (f"not.valid.{LocalModelFinder.__name__}", ValueError),
        (f"{LocalModelFinder.__module__}.FooBar", ValueError),
        (f"{LocalModelFinder.__module__}.alog", TypeError),
        (12345, TypeError),
    ],
)
def test_importable_factory_error_cases(params):
    """Make sure that all forms of bad value result in a ValueError or TypeError
    is raised
    """
    import_class_val, exc_type = params
    fact = ImportableFactory("TestFact")
    assert not fact._registered_types
    with pytest.raises(exc_type):
        fact.construct(
            {
                ImportableFactory.IMPORT_CLASS_KEY: import_class_val,
                ImportableFactory.TYPE_KEY: LocalModelFinder.name,
            }
        )



# Copyright The Caikit Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
This module helps us know when an rpc call is cancelled, and we need to stop or undo work
"""
# Standard
import threading

# Third Party
import grpc

# First Party
import alog

# Local
from caikit.runtime.types.caikit_runtime_exception import CaikitRuntimeException
from caikit.runtime.work_management.abortable_action import ActionAborter

log = alog.use_channel("CALL-ABORTER")


class RpcAborter(ActionAborter):
    """
    This class registers a callback with a grpc context, to be called in the event of rpc
    termination. Termination could be nominal (we returned a response) but we should have
    relinquished control anyway. The interesting case is when a client cancels a call or a deadline
    is hit, which could trigger this callback but will not interrupt the thread doing work.

    In order to actually interrupt threads doing the work, events can be registered with an
    instance of this class in order ton receive notification on RPC termination.

    IFF the RPC has been terminated, `must_abort` will return True.
    """

    def __init__(self, context: grpc.ServicerContext):
        # Create an event that we can use to check RPC termination
        self.is_terminated = False
        # Add an empty list for condition variables that will be notified on termination
        self.events = []

        callback_registered = context.add_callback(self.__rpc_terminated)

        # If we can't register the callback, it probably means the RPC has already been terminated.
        # We probably shouldn't be doing any more work.
        if not callback_registered:
            log.warning(
                "<RUN65620101W>",
                "Failed to register rpc termination callback, aborting rpc",
            )
            raise CaikitRuntimeException(
                grpc.StatusCode.ABORTED,
                "Could not register RPC callback, call has likely terminated.",
            )

    def must_abort(self):
        return self.is_terminated

    def add_event(self, event: threading.Event):
        self.events.append(event)

        # Sanity check: If we have already terminated, notify anything waiting on this condition
        if self.must_abort():
            event.set()

    def __rpc_terminated(self):
        # First set the flag so anybody waiting on us knows that gRPC wants us to abort work
        self.is_terminated = True

        # Then notify everybody waiting on us
        for event in self.events:
            event.set()



# Copyright The Caikit Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

# Standard
from typing import List

# Local
from caikit.core import DataObjectBase
from caikit.core.data_model import dataobject


@dataobject(package="text_sentiment.data_model")
class ClassInfo(DataObjectBase):
    """A single classification prediction."""

    class_name: str  # (required) Predicted relevant class name
    confidence: float  # (required) The confidence-like score of this prediction in [0, 1]


@dataobject(package="text_sentiment.data_model")
class ClassificationPrediction(DataObjectBase):
    """The result of a classification prediction."""

    classes: List[ClassInfo]



"""
This sets up global test configs when pytest starts
"""

# Standard
from contextlib import contextmanager
from typing import Callable, List, Union
from unittest.mock import patch
import copy
import json
import os
import sys
import tempfile
import uuid

# Third Party
import pytest

# First Party
import alog

# Local
from caikit import get_config
from caikit.core import MODEL_MANAGER
from caikit.core.registries import (
    module_backend_classes,
    module_backend_registry,
    module_backend_types,
    module_registry,
)
from caikit.core.toolkit import logging
import caikit

log = alog.use_channel("TEST-CONFTEST")

FIXTURES_DIR = os.path.join(
    os.path.dirname(__file__),
    "fixtures",
)

# Make sample_lib available for import
sys.path.append(FIXTURES_DIR)
# Local
import sample_lib

# Configure logging from the environment
logging.configure()


def random_test_id():
    return "test-any-model-" + _random_id()


@pytest.fixture(autouse=True, scope="session")
def test_environment():
    """The most important fixture: This runs caikit configuration with the base test config overrides"""
    test_config_path = os.path.join(FIXTURES_DIR, "config", "config.yml")
    caikit.configure(test_config_path)
    with tempfile.TemporaryDirectory() as workdir:
        # 🌶️ All tests need to clean up after themselves! The metering and
        # training output directories are the one thing easily creates leftover
        # files, so we explicitly update config here to point them at a temp dir
        with temp_config(
            {
                "runtime": {
                    "metering": {
                        "log_dir": os.path.join(workdir, "metering_logs"),
                    },
                    "training": {
                        "output_dir": os.path.join(workdir, "training_output"),
                    },
                }
            },
            "merge",
        ):
            # import the mock backend that is specified in the config
            # This is required to run any test that loads a model
            # Local
            from tests.core.helpers import MockBackend

            yield
    # No cleanup required...?


@pytest.fixture
def good_model_path() -> str:
    return os.path.join(FIXTURES_DIR, "models", "foo")


@pytest.fixture
def box_model_path() -> str:
    return os.path.join(FIXTURES_DIR, "models", "box")


@pytest.fixture
def streaming_model_path() -> str:
    return os.path.join(FIXTURES_DIR, "dummy_streaming_module")


@pytest.fixture
def other_good_model_path() -> str:
    return os.path.join(FIXTURES_DIR, "models", "bar")


# Sample data files for testing ###########################
@pytest.fixture
def data_stream_inputs() -> str:
    return os.path.join(FIXTURES_DIR, "data_stream_inputs")


@pytest.fixture
def sample_json_file(data_stream_inputs) -> str:
    return os.path.join(data_stream_inputs, "sample.json")


@pytest.fixture
def sample_json_collection(data_stream_inputs) -> str:
    return os.path.join(data_stream_inputs, "sample_json_collection")


@pytest.fixture
def sample_csv_file(data_stream_inputs) -> str:
    return os.path.join(data_stream_inputs, "sample_w_header.csv")


@pytest.fixture
def sample_csv_file_no_headers(data_stream_inputs) -> str:
    return os.path.join(data_stream_inputs, "sample.csv")


@pytest.fixture
def sample_csv_collection(data_stream_inputs) -> str:
    return os.path.join(data_stream_inputs, "sample_csv_collection")


@pytest.fixture
def sample_multipart_json(data_stream_inputs) -> str:
    return os.path.join(data_stream_inputs, "multi_part_json.txt")


@pytest.fixture
def sample_multipart_json_with_content_header(data_stream_inputs) -> str:
    return os.path.join(data_stream_inputs, "multi_part_json_with_content_header.txt")


@pytest.fixture
def sample_multipart_csv(data_stream_inputs) -> str:
    return os.path.join(data_stream_inputs, "multi_part_csv.txt")


@pytest.fixture
def sample_jsonl_file(data_stream_inputs) -> str:
    return os.path.join(data_stream_inputs, "sample.jsonl")


@pytest.fixture
def sample_jsonl_collection(data_stream_inputs) -> str:
    return os.path.join(data_stream_inputs, "sample_jsonl_collection")


@pytest.fixture
def jsonl_with_control_chars(data_stream_inputs) -> str:
    return os.path.join(data_stream_inputs, "control_chars.jsonl")


@pytest.fixture
def sample_text_file(data_stream_inputs) -> str:
    return os.path.join(data_stream_inputs, "sample.txt")


@pytest.fixture
def sample_text_collection(data_stream_inputs) -> str:
    return os.path.join(data_stream_inputs, "sample_txt_collection")


@contextmanager
def temp_config(config_overrides: dict, merge_strategy="override"):
    """Temporarily edit the caikit config in a mock context"""
    existing_config = copy.deepcopy(getattr(caikit.config.config, "_CONFIG"))
    # Patch out the internal config, starting with a fresh copy of the current config
    with patch.object(caikit.config.config, "_CONFIG", existing_config):
        # Patch the immutable view of the config as well
        # This is required otherwise the updated immutable view will persist after the test
        with patch.object(caikit.config.config, "_IMMUTABLE_CONFIG", None):
            # Run our config overrides inside the patch
            if config_overrides:
                config_overrides["merge_strategy"] = merge_strategy
                caikit.configure(config_dict=config_overrides)
            else:
                # or just slap some random uuids in there. Barf, but we need to call `.configure()`
                caikit.configure(config_dict={str(uuid.uuid4()): str(uuid.uuid4())})
            # Yield to the test with the new overridden config
            yield get_config()


@contextmanager
def set_use_subprocess(use_subprocess: bool):
    with temp_config(
        {
            "model_management": {
                "trainers": {
                    "default": {
                        "config": {
                            "use_subprocess": use_subprocess,
                        }
                    }
                }
            }
        },
        "merge",
    ):
        yield


# fixtures to optionally generate the protos for easier debugging
def pytest_addoption(parser):
    try:
        parser.addoption(
            "--render-protos",
            action="store_true",
            default=False,
            help="Render test protos for debug?",
        )
    except ValueError:
        pass


@pytest.fixture(scope="session")
def render_protos(request):
    return request.config.getoption("--render-protos")


@pytest.fixture
def sample_int_file() -> str:
    content = "[1,2,3,4,5,6,7,8,9,10]"
    with tempfile.NamedTemporaryFile(mode="w", suffix=".json") as handle:
        handle.write(content)
        handle.flush()
        yield handle.name


@pytest.fixture
def fixtures_dir():
    yield FIXTURES_DIR


@pytest.fixture
def modules_fixtures_dir(fixtures_dir):
    yield os.path.join(fixtures_dir, "modules")


@pytest.fixture
def toolkit_fixtures_dir(fixtures_dir):
    yield os.path.join(fixtures_dir, "toolkit")


@contextmanager
def backend_priority(backend_cfg: Union[List[dict], dict]):
    if isinstance(backend_cfg, dict):
        backend_cfg = [backend_cfg]
    with temp_config(
        {
            "model_management": {
                "finders": {"default": {"type": "LOCAL"}},
                "initializers": {
                    "default": {
                        "type": "LOCAL",
                        "config": {"backend_priority": backend_cfg},
                    }
                },
            }
        }
    ):
        yield


class TempFailWrapper:
    """Helper that can wrap a callable with a sequence of failures"""

    def __init__(
        self,
        func: Callable,
        num_failures: int = 1,
        exc: Exception = RuntimeError("Yikes"),
    ):
        self.func = func
        self.num_failures = num_failures
        self.exc = exc

    def __call__(self, *args, **kwargs):
        if self.num_failures:
            self.num_failures -= 1
            raise self.exc
        return self.func(*args, **kwargs)


# IMPLEMENTATION DETAILS ############################################################
def _random_id():
    return str(uuid.uuid4())


@pytest.fixture
def reset_backend_types():
    """Fixture that will reset the backend types if a test modifies them"""
    base_backend_types = {key: val for key, val in module_backend_types().items()}
    base_backend_classes = {key: val for key, val in module_backend_classes().items()}
    yield
    module_backend_types().clear()
    module_backend_types().update(base_backend_types)
    module_backend_classes().clear()
    module_backend_classes().update(base_backend_classes)


@pytest.fixture
def reset_module_backend_registry():
    """Fixture that will reset the module distribution registry if a test modifies them"""
    # NB: Deepcopy here because the values are nested dicts that will be persisted by reference
    orig_module_backend_registry = copy.deepcopy(module_backend_registry())
    yield
    module_backend_registry().clear()
    module_backend_registry().update(orig_module_backend_registry)


@pytest.fixture
def reset_module_registry():
    """Fixture that will reset caikit.core module registry if a test modifies it"""
    orig_module_registry = {key: val for key, val in module_registry().items()}
    yield
    module_registry().clear()
    module_registry().update(orig_module_registry)


@pytest.fixture
def reset_model_manager():
    prev_finders = MODEL_MANAGER._finders
    prev_initializers = MODEL_MANAGER._initializers
    prev_trainers = MODEL_MANAGER._trainers
    MODEL_MANAGER._finders = {}
    MODEL_MANAGER._initializers = {}
    MODEL_MANAGER._trainers = {}
    yield
    MODEL_MANAGER._finders = prev_finders
    MODEL_MANAGER._initializers = prev_initializers
    MODEL_MANAGER._trainers = prev_trainers


@pytest.fixture
def reset_globals(
    reset_backend_types,
    reset_model_manager,
    reset_module_backend_registry,
    reset_module_registry,
):
    """Fixture that will reset the backend types and module registries if a test modifies them"""



# Copyright The Caikit Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""
Caikit Core Exception enum used for reporting Exception status raised in caikit core
"""

# Standard
from enum import Enum


class CaikitCoreStatusCode(Enum):
    NOT_FOUND = 1
    INVALID_ARGUMENT = 2
    CONNECTION_ERROR = 3
    UNAUTHORIZED = 4
    FORBIDDEN = 5
    UNKNOWN = 6
    FATAL = 7


class CaikitCoreException(Exception):
    status_code: CaikitCoreStatusCode
    message: str

    def __init__(self, status_code: CaikitCoreStatusCode, message: str) -> None:
        self.status_code = status_code
        self.message = message



# Copyright The Caikit Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


"""Common data model containing all data structures that are passed in and out of modules.
"""

# Local
from . import base, data_backends, enums, producer, protobufs
from .base import DataBase
from .dataobject import (
    CAIKIT_DATA_MODEL,
    DataObjectBase,
    dataobject,
    render_dataobject_protos,
)
from .enums import *
from .producer import PACKAGE_COMMON, ProducerId
from .streams import data_stream
from .streams.data_stream import *
from .training_status import TrainingStatus



# Copyright The Caikit Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""This module defines the @schema decorator which can be used to declare data
model objects inline without manually defining the protobufs representation
"""


# Standard
from enum import Enum
from typing import (
    Any,
    Callable,
    Dict,
    Iterable,
    List,
    Optional,
    Type,
    TypeVar,
    Union,
    get_args,
    get_origin,
)
import dataclasses

# Third Party
from google.protobuf import descriptor as _descriptor
from google.protobuf import message as _message
from google.protobuf import struct_pb2
from google.protobuf.internal.enum_type_wrapper import EnumTypeWrapper
import numpy as np

# First Party
from py_to_proto.dataclass_to_proto import PY_TO_PROTO_TYPES, DataclassConverter
import alog
import py_to_proto

# Local
from ..exceptions import error_handler
from . import enums
from .base import DataBase, _DataBaseMetaClass
from .json_dict import JsonDict

## Globals #####################################################################

log = alog.use_channel("SCHEMA")
error = error_handler.get(log)

# Type mapping for type hints in @dataobject classes
DATAOBJECT_PY_TO_PROTO_TYPES = {
    JsonDict: struct_pb2.Struct,
    np.int32: _descriptor.FieldDescriptor.TYPE_INT32,
    np.int64: _descriptor.FieldDescriptor.TYPE_INT64,
    np.uint32: _descriptor.FieldDescriptor.TYPE_UINT32,
    np.uint64: _descriptor.FieldDescriptor.TYPE_UINT64,
    np.float32: _descriptor.FieldDescriptor.TYPE_FLOAT,
    np.float64: _descriptor.FieldDescriptor.TYPE_DOUBLE,
    **PY_TO_PROTO_TYPES,
}

# Common package prefix
CAIKIT_DATA_MODEL = "caikit_data_model"

# Registry of auto-generated protos so that they can be rendered to .proto
_AUTO_GEN_PROTO_CLASSES = []

# Special attribute used to indicate which defaults are user provided
_USER_DEFINED_DEFAULTS = "__user_defined_defaults__"

## Public ######################################################################


class _DataObjectBaseMetaClass(_DataBaseMetaClass):
    """This metaclass is used for the DataObject base class so that all data
    objects can delay the creation of their proto class until after the
    metaclass has been instantiated.
    """

    def __new__(mcs, name, bases, attrs):
        """When instantiating a new DataObject class, the proto class will not
        yet have been generated, but the set of fields will be known since the
        class will be the raw input representation of a @dataclass
        """

        # Get the annotations that will go into the dataclass
        if name != "DataObjectBase":
            field_names = attrs.get("__annotations__")
            parent_dataobjects = [
                base for base in bases if isinstance(base, _DataBaseMetaClass)
            ]
            field_name_sets = [base.fields for base in parent_dataobjects]
            if field_names is not None:
                field_name_sets = [field_names] + field_name_sets

            # We need at least one field name set!
            if not field_name_sets:
                raise TypeError(
                    "All DataObjectBase classes must follow dataclass syntax"
                )

            # Flatten the field names
            field_names = {
                field_name
                for field_name_set in field_name_sets
                for field_name in field_name_set
            }
            attrs[_DataBaseMetaClass._FWD_DECL_FIELDS] = field_names

        # Delegate to the base metaclass
        return super().__new__(mcs, name, bases, attrs)


class DataObjectBase(DataBase, metaclass=_DataObjectBaseMetaClass):
    """A DataObject is a data model class that is backed by a @dataclass.

    Data model classes that use the @dataobject decorator must derive from this
    base class.
    """


_DataObjectBaseT = TypeVar("_DataObjectBaseT", bound=DataObjectBase)


def dataobject(*args, **kwargs) -> Callable[[_DataObjectBaseT], _DataObjectBaseT]:
    """The @dataobject decorator can be used to define a Data Model object's
    schema inline with the definition of the python class rather than needing to
    bind to a pre-compiled protobufs class. For example:

    @dataobject("foo.bar")
    class MyDataObject(DataObjectBase):
        '''My Custom Data Object'''
        foo: str
        bar: int

    NOTE: The wrapped class must NOT inherit directly from DataBase. That
        inheritance will be added by this decorator, but if it is written
        directly, the metaclass that links protobufs to the class will be called
        before this decorator can auto-gen the protobufs class.

    The `dataobject` decorator will not provide tools with enough information
    to perform type completion for constructions in an IDE, or static
    typechecking.  In order to have that, the `dataclass` decorator
    may optionally be added, with the slight overhead of wasted effort in
    creating the "standard" __init__ function which then gets re-done by
    @dataobject.  The `dataclass` must follow the `dataobject` decorator.  For example:

    @dataobject("foo.bar")
    @dataclass
    class MyDataObject(DataObjectBase):
        '''My Custom Data Object'''
        foo: str
        bar: int

    Kwargs:
        package:  str
            The package name to use for the generated protobufs class

    Returns:
        decorator:  Callable[[Type], Type[DataBase]]
            The decorator function that will wrap the given class
    """

    def decorator(cls: _DataObjectBaseT) -> _DataObjectBaseT:
        # Make sure that the wrapped class does NOT inherit from DataBase
        error.value_check(
            "<COR95184230E>",
            issubclass(cls, (DataObjectBase, Enum)),
            "{} must inherit from DataObjectBase/Enum when using @dataobject",
            cls.__name__,
        )

        # Add the package to the kwargs
        kwargs.setdefault("package", package)

        # If it's not an enum, fill in any missing field defaults as None
        # and make sure it's a dataclass
        if not issubclass(cls, Enum):
            log.debug2("Wrapping data class %s", cls)
            user_defined_defaults = {}
            for annotation in getattr(cls, "__annotations__", {}):
                user_defined_default = getattr(cls, annotation, dataclasses.MISSING)
                if user_defined_default == dataclasses.MISSING:
                    log.debug3("Filling in None default for %s.%s", cls, annotation)
                    setattr(cls, annotation, None)
                else:
                    user_defined_defaults[annotation] = user_defined_default
            # If the current __init__ is auto-generated by dataclass, remove
            # it so that a new one is created with the new defaults
            if _has_dataclass_init(cls):
                log.debug3("Resetting default dataclass init")
                delattr(cls, "__init__")
            cls = dataclasses.dataclass(repr=False)(cls)
            setattr(cls, _USER_DEFINED_DEFAULTS, user_defined_defaults)

        descriptor = _dataobject_to_proto(dataclass_=cls, **kwargs)

        # Create the message class from the dataclass
        proto_class = py_to_proto.descriptor_to_message_class(descriptor)
        _AUTO_GEN_PROTO_CLASSES.append(proto_class)

        # Add enums to the global enums module
        for enum_class in _get_all_enums(proto_class):
            log.debug2("Importing enum [%s]", enum_class.DESCRIPTOR.name)
            enums.import_enum(enum_class)

        # Declare the merged class that binds DataBase to the wrapped class with
        # this generated proto class
        if not isinstance(proto_class, EnumTypeWrapper):
            setattr(cls, "_proto_class", proto_class)
            cls = _make_data_model_class(proto_class, cls)

            # If this was a default-generated dataclass __init__ and there are
            # any oneofs, we need to augment the __init__ to support kwargs for
            # the individual fields
            if _has_dataclass_init(cls) and cls._fields_oneofs_map:
                setattr(cls, "__init__", _make_oneof_init(cls))

        else:
            enums.import_enum(proto_class, cls)
            setattr(cls, "_proto_enum", proto_class)

        # Return the decorated class
        return cls

    # If called without the function invocation, fill in the default argument
    if args and callable(args[0]):
        assert not kwargs, "This shouldn't happen!"
        package = CAIKIT_DATA_MODEL
        return decorator(args[0])

    # Pull the package as an arg or a keyword arg
    if args:
        package = args[0]
        if "package" in kwargs:
            raise TypeError("Got multiple values for argument 'package'")
    else:
        package = kwargs.get("package", CAIKIT_DATA_MODEL)
    return decorator


def render_dataobject_protos(interfaces_dir: str):
    """Write out protobufs files for all proto classes generated from dataobjects
    to the target interfaces directory

    Args:
        interfaces_dir (str): The target directory (must already exist)
    """
    for proto_class in _AUTO_GEN_PROTO_CLASSES:
        proto_class.write_proto_file(interfaces_dir)


def make_dataobject(
    *,
    name: str,
    annotations: Dict[str, type],
    bases: Optional[Iterable[type]] = None,
    attrs: Optional[Dict[str, Any]] = None,
    proto_name: Optional[str] = None,
    **kwargs,
) -> _DataObjectBaseMetaClass:
    """Factory function for creating net-new dataobject classes

    WARNING: This is a power-user feature that should be used with caution since
        dynamically generated dataobject classes have portability issues due to
        the use of global registries.

    Kwargs:
        name (str): The name of the class to create
        annotations (Dict[str, type]): The type annotations for the class
        bases (Optional[Iterable[type]]): Additional base classes beyond
            DataObjectBase
        attrs (Optional[Dict[str, Any]]): Additional class attributes beyond
            __annotations__
        proto_name (Optional[str]): Alternate name to use for the name of
            protobuf message

    Returns:
        dataobject_class (_DataObjectBaseMetaClass): Programmatically created
            class derived from DataObjectBase with the given name and
            annotations
    """
    bases = (DataObjectBase,) + tuple(bases or ())
    attrs = {
        "__annotations__": annotations,
        **(attrs or {}),
    }
    if proto_name is not None:
        kwargs["name"] = proto_name
    return dataobject(**kwargs)(
        _DataObjectBaseMetaClass.__new__(
            _DataObjectBaseMetaClass,
            name=name,
            bases=bases,
            attrs=attrs,
        )
    )


## Implementation Details ######################################################


def _dataobject_to_proto(*args, **kwargs):
    kwargs.setdefault("type_mapping", DATAOBJECT_PY_TO_PROTO_TYPES)
    return _DataobjectConverter(*args, **kwargs).descriptor


class _DataobjectConverter(DataclassConverter):
    """Augment the dataclass converter to be able to pull descriptors from
    existing data objects
    """

    def get_concrete_type(self, entry: Any) -> Any:
        """Also include data model classes and enums as concrete types"""
        unwrapped = self._resolve_wrapped_type(entry)
        if (
            isinstance(unwrapped, type)
            and issubclass(unwrapped, DataBase)
            and entry.get_proto_class() is not None
        ) or hasattr(unwrapped, "_proto_enum"):
            return entry
        return super().get_concrete_type(entry)

    def get_descriptor(self, entry: Any) -> Any:
        """Unpack data model classes and enums to their descriptors"""
        entry = self._resolve_wrapped_type(entry)
        if isinstance(entry, type) and issubclass(entry, DataBase):
            return entry.get_proto_class().DESCRIPTOR
        proto_enum = getattr(entry, "_proto_enum", None)
        if proto_enum is not None:
            return proto_enum.DESCRIPTOR
        return super().get_descriptor(entry)

    def get_optional_field_names(self, entry: Any) -> List[str]:
        """Get the names of any fields which are optional. This will be any
        field that has a user-defined default or is marked as Optional[]
        """
        optional_fields = list(getattr(entry, _USER_DEFINED_DEFAULTS, {}))
        for field_name, field in entry.__dataclass_fields__.items():
            if (
                field_name not in optional_fields
                and self._is_python_optional(field.type) is not None
            ):
                optional_fields.append(field_name)
        return optional_fields

    @staticmethod
    def _is_python_optional(entry: Any) -> Any:
        """Detect if this type is a python optional"""
        if get_origin(entry) is Union:
            args = get_args(entry)
            return type(None) in args


def _get_all_enums(
    proto_class: Union[_message.Message, EnumTypeWrapper],
) -> List[EnumTypeWrapper]:
    """Given a generated proto class, recursively extract all enums"""
    all_enums = []
    if isinstance(proto_class, EnumTypeWrapper):
        all_enums.append(proto_class)
    else:
        for enum_descriptor in proto_class.DESCRIPTOR.enum_types:
            all_enums.append(getattr(proto_class, enum_descriptor.name))
        for nested_proto_descriptor in proto_class.DESCRIPTOR.nested_types:
            all_enums.extend(
                _get_all_enums(getattr(proto_class, nested_proto_descriptor.name))
            )

    return all_enums


def _make_data_model_class(proto_class: Type[_message.Message], cls):
    if issubclass(cls, DataObjectBase):
        _DataBaseMetaClass.parse_proto_descriptor(cls)

    # Recursively make all nested message wrappers
    for nested_message_descriptor in proto_class.DESCRIPTOR.nested_types:
        nested_message_name = nested_message_descriptor.name
        nested_proto_class = getattr(proto_class, nested_message_name)
        setattr(
            cls,
            nested_message_name,
            _make_data_model_class(
                nested_proto_class,
                _DataBaseMetaClass.__new__(
                    _DataBaseMetaClass,
                    name=nested_message_name,
                    bases=(DataBase,),
                    attrs={"_proto_class": getattr(proto_class, nested_message_name)},
                ),
            ),
        )
    for nested_enum_descriptor in proto_class.DESCRIPTOR.enum_types:
        setattr(
            cls,
            nested_enum_descriptor.name,
            getattr(enums, nested_enum_descriptor.name),
        )

    return cls


def _make_oneof_init(cls):
    """Helper to augment a defaulted dataclass __init__ to support kwargs for
    oneof fields
    """
    original_init = cls.__init__
    fields_to_oneofs = cls._fields_to_oneof
    oneofs_to_fields = cls._fields_oneofs_map

    def __init__(self, *args, **kwargs):
        new_kwargs = {}
        to_remove = []
        which_oneof = {}
        for field_name, val in kwargs.items():
            if oneof_name := fields_to_oneofs.get(field_name):
                oneof_pos_idx = list(cls.__dataclass_fields__.keys()).index(oneof_name)
                has_pos_val = len(args) > oneof_pos_idx
                if has_pos_val:
                    error(
                        "<COR09282193E>",
                        TypeError(
                            "Received conflicting oneof args/kwargs for {}/{}".format(
                                oneof_name,
                                field_name,
                            )
                        ),
                    )

                other_oneof_fields = (
                    field
                    for field in [oneof_name] + oneofs_to_fields[oneof_name]
                    if field != field_name
                )
                if any(field in kwargs for field in other_oneof_fields):
                    error(
                        "<COR59933157E>",
                        TypeError(
                            "Received multiple keyword arguments for oneof {}".format(
                                oneof_name,
                            )
                        ),
                    )
                new_kwargs[oneof_name] = val
                to_remove.append(field_name)
                which_oneof[oneof_name] = field_name

        for kwarg in to_remove:
            del kwargs[kwarg]
        kwargs.update(new_kwargs)
        original_init(self, *args, **kwargs)
        # noinspection PyProtectedMember
        setattr(self, _DataBaseMetaClass._WHICH_ONEOF_ATTR, which_oneof)

    return __init__


def _has_dataclass_init(cls) -> bool:
    """When the dataclass decorator adds an __init__ to a class, it adds
    __annotations__ to the init function itself. This function uses that fact to
    detect if the class's __init__ function was generated by @dataclass
    """
    return bool(getattr(cls.__init__, "__annotations__", None)) and not any(
        cls.__init__ is base.__init__ for base in cls.__bases__
    )



# Copyright The Caikit Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""
This module holds the implementation of caikit's primary HTTP server entrypoint.
The server is responsible for binding caikit workloads to a consistent REST/SSE
API based on the task definitions available at boot.
"""
# Standard
from contextlib import contextmanager
from dataclasses import dataclass
from functools import partial
from typing import Any, Dict, Iterable, Optional, Type, get_args
import asyncio
import json
import os
import re
import ssl
import tempfile
import threading
import time

# Third Party
from fastapi import FastAPI, Request, Response, status
from fastapi.encoders import jsonable_encoder
from fastapi.exceptions import ResponseValidationError
from fastapi.responses import JSONResponse, PlainTextResponse
from grpc import StatusCode
from sse_starlette import EventSourceResponse, ServerSentEvent
import pydantic
import uvicorn

# First Party
from py_to_proto.dataclass_to_proto import get_origin  # Imported here for 3.8 compat
import aconfig
import alog

# Local
from .pydantic_wrapper import dataobject_to_pydantic, pydantic_to_dataobject
from caikit.config import get_config
from caikit.core.data_model import DataBase
from caikit.core.data_model.dataobject import make_dataobject
from caikit.core.toolkit.sync_to_async import async_wrap_iter
from caikit.runtime.server_base import RuntimeServerBase
from caikit.runtime.service_factory import ServicePackage
from caikit.runtime.service_generation.rpcs import (
    CaikitRPCBase,
    ModuleClassTrainRPC,
    TaskPredictRPC,
)
from caikit.runtime.servicers.global_predict_servicer import GlobalPredictServicer
from caikit.runtime.servicers.global_train_servicer import GlobalTrainServicer
from caikit.runtime.types.caikit_runtime_exception import CaikitRuntimeException

## Globals #####################################################################

log = alog.use_channel("SERVR-HTTP")


# Mapping from GRPC codes to their corresponding HTTP codes
# pylint: disable=line-too-long
# CITE: https://chromium.googlesource.com/external/github.com/grpc/grpc/+/refs/tags/v1.21.4-pre1/doc/statuscodes.md
GRPC_CODE_TO_HTTP = {
    StatusCode.OK: 200,
    StatusCode.INVALID_ARGUMENT: 400,
    StatusCode.FAILED_PRECONDITION: 400,
    StatusCode.OUT_OF_RANGE: 400,
    StatusCode.UNAUTHENTICATED: 401,
    StatusCode.PERMISSION_DENIED: 403,
    StatusCode.NOT_FOUND: 404,
    StatusCode.ALREADY_EXISTS: 409,
    StatusCode.ABORTED: 409,
    StatusCode.RESOURCE_EXHAUSTED: 429,
    StatusCode.CANCELLED: 499,
    StatusCode.UNKNOWN: 500,
    StatusCode.DATA_LOSS: 500,
    StatusCode.UNIMPLEMENTED: 501,
    StatusCode.UNAVAILABLE: 501,
    StatusCode.DEADLINE_EXCEEDED: 504,
}


# These keys are used to define the logical sections of the request and response
# data structures.
REQUIRED_INPUTS_KEY = "inputs"
OPTIONAL_INPUTS_KEY = "parameters"

# Endpoint to use for health checks
HEALTH_ENDPOINT = "/health"

# Small dataclass for consolidating TLS files
@dataclass
class _TlsFiles:
    server_cert: Optional[str] = None
    server_key: Optional[str] = None
    client_cert: Optional[str] = None


## RuntimeHTTPServer ###########################################################


class RuntimeHTTPServer(RuntimeServerBase):
    """An implementation of a FastAPI server that serves caikit runtimes"""

    ###############
    ## Interface ##
    ###############

    def __init__(self, tls_config_override: Optional[aconfig.Config] = None):
        super().__init__(get_config().runtime.http.port, tls_config_override)

        self.app = FastAPI()

        # Response validation
        @self.app.exception_handler(ResponseValidationError)
        async def validation_exception_handler(_, exc: ResponseValidationError):
            return JSONResponse(
                status_code=status.HTTP_500_INTERNAL_SERVER_ERROR,
                content=jsonable_encoder({"detail": exc.errors(), "body": exc.body}),
            )

        # Start metrics server
        RuntimeServerBase._start_metrics_server()

        # Placeholders for global servicers
        self.global_predict_servicer = None
        self.global_train_servicer = None

        # Set up inference if enabled
        if self.enable_inference:
            log.info("<RUN77183426I>", "Enabling HTTP inference service")
            self.global_predict_servicer = GlobalPredictServicer(self.inference_service)
            self._bind_routes(self.inference_service)

        # Set up training if enabled
        if self.enable_training:
            log.info("<RUN77183427I>", "Enabling HTTP training service")
            self.global_train_servicer = GlobalTrainServicer(self.training_service)
            self._bind_routes(self.training_service)

        # Add the health endpoint
        self.app.get(HEALTH_ENDPOINT, response_class=PlainTextResponse)(
            self._health_check
        )

        # Parse TLS configuration
        # If any of the TLS values are not files, we assume that they're inline
        # content. The python SslContext only takes files to load, so we use a
        # temporary directory just long enough to load the config files.
        with self._tls_files() as tls_files:
            tls_kwargs = {}
            if tls_files.server_key and tls_files.server_cert:
                log.info("<RUN10001905I>", "Running with TLS")
                tls_kwargs["ssl_keyfile"] = tls_files.server_key
                tls_kwargs["ssl_certfile"] = tls_files.server_cert
                if tls_files.client_cert:
                    log.info("<RUN10001809I>", "Running with mutual TLS")
                    tls_kwargs["ssl_ca_certs"] = tls_files.client_cert
                    tls_kwargs["ssl_cert_reqs"] = ssl.CERT_REQUIRED
            else:
                log.info("<RUN10539515I>", "Running INSECURE")

            # Start the server with a timeout_graceful_shutdown
            # if not set in config, this is None and unvicorn accepts None or number of seconds
            unvicorn_timeout_graceful_shutdown = (
                get_config().runtime.http.server_shutdown_grace_period_seconds
            )
            config = uvicorn.Config(
                self.app,
                host="0.0.0.0",
                port=self.port,
                log_level=None,
                log_config=None,
                timeout_graceful_shutdown=unvicorn_timeout_graceful_shutdown,
                **tls_kwargs,
            )
            # Make sure the config loads TLS files here so they can safely be
            # deleted if they're ephemeral
            config.load()

        # Start the server with the loaded config
        self.server = uvicorn.Server(config=config)

        # Patch the exit handler to call this server's stop
        original_handler = self.server.handle_exit

        def shutdown_wrapper(*args, **kwargs):
            original_handler(*args, **kwargs)
            self.stop()

        self.server.handle_exit = shutdown_wrapper

        # Placeholder for thread when running without blocking
        self._uvicorn_server_thread = None

    def __del__(self):
        config = get_config()
        if config and config.runtime.metering.enabled and self.global_predict_servicer:
            self.global_predict_servicer.stop_metering()

    def start(self, blocking: bool = True):
        """Boot the http server. Can be non-blocking, or block until shutdown

        Args:
            blocking (boolean): Whether to block until shutdown
        """
        if blocking:
            self.server.run()
        else:
            self.run_in_thread()

    def stop(self):
        """Stop the server, with an optional grace period.

        Args:
            grace_period_seconds (Union[float, int]): Grace period for service shutdown.
                Defaults to application config
        """
        self.server.should_exit = True
        if (
            self._uvicorn_server_thread is not None
            and self._uvicorn_server_thread.is_alive()
        ):
            self._uvicorn_server_thread.join()

        # Ensure we flush out any remaining billing metrics and stop metering
        if self.config.runtime.metering.enabled and self.global_predict_servicer:
            self.global_predict_servicer.stop_metering()

        # Shut down the model manager's model polling if enabled
        self._shut_down_model_manager()

    def run_in_thread(self):
        self._uvicorn_server_thread = threading.Thread(target=self.server.run)
        self._uvicorn_server_thread.start()
        while not self.server.started:
            time.sleep(1e-3)
        log.info("HTTP Server is running in thread")

    ##########
    ## Impl ##
    ##########
    def _bind_routes(self, service: ServicePackage):
        """Bind all rpcs as routes to the given app"""
        for rpc in service.caikit_rpcs.values():
            rpc_info = rpc.create_rpc_json("")
            if isinstance(rpc, TaskPredictRPC):
                if hasattr(rpc, "input_streaming") and rpc.input_streaming:
                    # Skipping the binding of this route since we don't have support
                    log.info(
                        "No support for input streaming on REST Server yet! Skipping this rpc %s with input type %s",
                        rpc_info["name"],
                        rpc_info["input_type"],
                    )
                    continue
                if hasattr(rpc, "output_streaming") and rpc.output_streaming:
                    self._add_unary_input_stream_output_handler(rpc)
                else:
                    self._add_unary_input_unary_output_handler(rpc)
            elif isinstance(rpc, ModuleClassTrainRPC):
                self._train_add_unary_input_unary_output_handler(rpc)

    def _get_request_params(
        self, rpc: CaikitRPCBase, request: Type[pydantic.BaseModel]
    ) -> Dict[str, Any]:
        """get the request params based on the RPC's req params, also
        convert to DM objects"""
        request_kwargs = dict(request)
        input_name = None
        required_params = None
        if isinstance(rpc, TaskPredictRPC):
            required_params = rpc.task.get_required_parameters(rpc.input_streaming)
        # handle required param input name
        if required_params and len(required_params) == 1:
            input_name = list(required_params.keys())[0]
        # flatten inputs and params into a dict
        # would have been useful to call dataobject.to_dict()
        # but unfortunately we now have converted pydantic objects
        combined_dict = {}
        for field, value in request_kwargs.items():
            if value:
                if field == REQUIRED_INPUTS_KEY and input_name:
                    combined_dict[input_name] = value
                else:
                    combined_dict.update(**dict(request_kwargs[field]))
        # remove non-none items
        request_params = {k: v for k, v in combined_dict.items() if v is not None}
        # convert pydantic objects to our DM objects
        for param_name, param_value in request_params.items():
            if issubclass(type(param_value), pydantic.BaseModel):
                request_params[param_name] = pydantic_to_dataobject(param_value)
        return request_params

    def _train_add_unary_input_unary_output_handler(self, rpc: CaikitRPCBase):
        """Add a unary:unary request handler for this RPC signature"""
        pydantic_request = dataobject_to_pydantic(
            DataBase.get_class_for_name(rpc.request.name)
        )
        pydantic_response = dataobject_to_pydantic(self._get_response_dataobject(rpc))

        @self.app.post(self._get_route(rpc), response_model=pydantic_response)
        # pylint: disable=unused-argument
        async def _handler(request: pydantic_request, context: Request) -> Response:
            log.debug("In unary handler for %s", rpc.name)
            loop = asyncio.get_running_loop()

            # build request DM object
            http_request_dm_object = pydantic_to_dataobject(request)

            try:
                call = partial(
                    self.global_train_servicer.run_training_job,
                    request=http_request_dm_object.to_proto(),
                    module=rpc.clz,
                    training_output_dir=None,  # pass None so that GTS picks up the config one # TODO: double-check?
                    # context=context,
                    wait=True,
                )
                result = await loop.run_in_executor(None, call)
                return Response(content=result.to_json(), media_type="application/json")

            except CaikitRuntimeException as err:
                error_code = GRPC_CODE_TO_HTTP.get(err.status_code, 500)
                error_content = {
                    "details": err.message,
                    "code": error_code,
                    "id": err.id,
                }
            except Exception as err:  # pylint: disable=broad-exception-caught
                error_code = 500
                error_content = {
                    "details": f"Unhandled exception: {str(err)}",
                    "code": error_code,
                    "id": None,
                }
                log.error("<RUN51881106E>", err, exc_info=True)
            return Response(content=json.dumps(error_content), status_code=error_code)

    def _add_unary_input_unary_output_handler(self, rpc: CaikitRPCBase):
        """Add a unary:unary request handler for this RPC signature"""
        pydantic_request = dataobject_to_pydantic(self._get_request_dataobject(rpc))
        pydantic_response = dataobject_to_pydantic(self._get_response_dataobject(rpc))

        @self.app.post(self._get_route(rpc), response_model=pydantic_response)
        # pylint: disable=unused-argument
        async def _handler(
            model_id: str, request: pydantic_request, context: Request
        ) -> Response:
            log.debug("In unary handler for %s for model %s", rpc.name, model_id)
            loop = asyncio.get_running_loop()

            request_params = self._get_request_params(rpc, request)

            log.debug4("Sending request %s to model id %s", request_params, model_id)
            try:
                model = self.global_predict_servicer._model_manager.retrieve_model(
                    model_id
                )

                # TODO: use `async_wrap_*`?
                call = partial(
                    self.global_predict_servicer.predict_model,
                    model_id=model_id,
                    request_name=rpc.request.name,
                    inference_func_name=model.get_inference_signature(
                        output_streaming=False, input_streaming=False
                    ).method_name,
                    **request_params,
                )
                result = await loop.run_in_executor(None, call)
                log.debug4("Response from model %s is %s", model_id, result)
                return Response(content=result.to_json(), media_type="application/json")

            except CaikitRuntimeException as err:
                error_code = GRPC_CODE_TO_HTTP.get(err.status_code, 500)
                error_content = {
                    "details": err.message,
                    "code": error_code,
                    "id": err.id,
                }
            except Exception as err:  # pylint: disable=broad-exception-caught
                error_code = 500
                error_content = {
                    "details": f"Unhandled exception: {str(err)}",
                    "code": error_code,
                    "id": None,
                }
                log.error("<RUN51881106E>", err, exc_info=True)
            return Response(content=json.dumps(error_content), status_code=error_code)

    def _add_unary_input_stream_output_handler(self, rpc: CaikitRPCBase):
        pydantic_request = dataobject_to_pydantic(self._get_request_dataobject(rpc))
        pydantic_response = dataobject_to_pydantic(self._get_response_dataobject(rpc))

        # pylint: disable=unused-argument
        @self.app.post(self._get_route(rpc), response_model=pydantic_response)
        async def _handler(
            model_id: str, request: pydantic_request, context: Request
        ) -> EventSourceResponse:
            log.debug("In streaming handler for %s", rpc.name)

            request_params = self._get_request_params(rpc, request)
            log.debug4("Sending request %s to model id %s", request_params, model_id)

            async def _generator() -> pydantic_response:
                try:
                    model = self.global_predict_servicer._model_manager.retrieve_model(
                        model_id
                    )

                    log.debug("In stream generator for %s", rpc.name)
                    async for result in async_wrap_iter(
                        self.global_predict_servicer.predict_model(
                            model_id=model_id,
                            request_name=rpc.request.name,
                            inference_func_name=model.get_inference_signature(
                                output_streaming=True, input_streaming=False
                            ).method_name,
                            **request_params,
                        )
                    ):
                        yield result
                    return
                except CaikitRuntimeException as err:
                    error_code = GRPC_CODE_TO_HTTP.get(err.status_code, 500)
                    error_content = {
                        "details": err.message,
                        "code": error_code,
                        "id": err.id,
                    }
                except Exception as err:  # pylint: disable=broad-exception-caught
                    error_code = 500
                    error_content = {
                        "details": f"Unhandled exception: {str(err)}",
                        "code": error_code,
                        "id": None,
                    }
                    log.error("<RUN51881106E>", err, exc_info=True)

                # If an error occurs, yield an error response and terminate
                yield ServerSentEvent(data=json.dumps(error_content))

            return EventSourceResponse(_generator())

    def _get_route(self, rpc: CaikitRPCBase) -> str:
        """Get the REST route for this rpc"""
        if rpc.name.endswith("Predict"):
            task_name = re.sub(
                r"(?<!^)(?=[A-Z])",
                "-",
                re.sub("Task$", "", re.sub("Predict$", "", rpc.name)),
            ).lower()
            route = "/".join(
                [self.config.runtime.http.route_prefix, "{model_id}", "task", task_name]
            )
            if route[0] != "/":
                route = "/" + route
            return route
        if rpc.name.endswith("Train"):
            route = "/".join([self.config.runtime.http.route_prefix, rpc.name])
            if route[0] != "/":
                route = "/" + route
            return route
        raise NotImplementedError("No support for train rpcs yet!")

    def _get_request_dataobject(self, rpc: CaikitRPCBase) -> Type[DataBase]:
        """Get the dataobject request for the given rpc"""
        is_inference_rpc = hasattr(rpc, "task")
        if is_inference_rpc:
            required_params = rpc.task.get_required_parameters(rpc.input_streaming)
        else:  # train
            required_params = {
                entry[1]: entry[0]
                for entry in rpc.request.triples
                if entry[1] not in rpc.request.default_map
            }
        optional_params = {
            entry[1]: entry[0]
            for entry in rpc.request.triples
            if entry[1] not in required_params
        }

        inputs_type = None
        if is_inference_rpc:
            pkg_name = f"caikit.http.{rpc.task.__name__}"
        else:
            pkg_name = f"caikit.http.{rpc.name}"

        # Create a bundled sub-message for required parameters for multiple params
        if len(required_params) > 1:
            log.debug3("Using structured inputs type for %s", pkg_name)
            inputs_type = make_dataobject(
                name=f"{rpc.request.name}Inputs",
                annotations=required_params,
                package=pkg_name,
            )
        elif required_params:
            inputs_type = list(required_params.values())[0]
            log.debug3(
                "Using single inputs type for task %s: %s", pkg_name, inputs_type
            )

        # Always create a bundled sub-message for optional parameters
        parameters_type = None
        if optional_params:
            parameters_type = make_dataobject(
                name=f"{rpc.request.name}Parameters",
                annotations=optional_params,
                package=pkg_name,
            )

        # Create the top-level request message
        request_annotations = {}
        if inputs_type:
            request_annotations[REQUIRED_INPUTS_KEY] = inputs_type
        if parameters_type:
            request_annotations[OPTIONAL_INPUTS_KEY] = parameters_type
        request_message = make_dataobject(
            name=f"{rpc.request.name}HttpRequest",
            annotations=request_annotations,
            package=pkg_name,
        )

        return request_message

    @staticmethod
    def _get_response_dataobject(rpc: CaikitRPCBase) -> Type[DataBase]:
        """Get the dataobject response for the given rpc"""
        origin = get_origin(rpc.return_type)
        args = get_args(rpc.return_type)
        if isinstance(origin, type) and issubclass(origin, Iterable):
            assert args and len(args) == 1
            dm_obj = args[0]
        else:
            dm_obj = rpc.return_type
        assert isinstance(dm_obj, type) and issubclass(dm_obj, DataBase)
        return dm_obj

    @staticmethod
    def _health_check() -> str:
        log.debug4("Server healthy")
        return "OK"

    @contextmanager
    def _tls_files(self) -> _TlsFiles:
        """This contextmanager ensures that the tls config values are files on
        disk since SslContext requires files

        Returns:
            tls_files (_TlsFiles): The set of configured TLS files
        """
        tls_cfg = self.tls_config or {}
        tls_opts = {
            "server_key": tls_cfg.get("server", {}).get("key", ""),
            "server_cert": tls_cfg.get("server", {}).get("cert", ""),
            "client_cert": tls_cfg.get("client", {}).get("cert", ""),
        }
        non_file_opts = {
            key: val for key, val in tls_opts.items() if val and not os.path.isfile(val)
        }
        if not non_file_opts:
            log.debug3("No TLS inline files configured")
            yield _TlsFiles(**tls_opts)
            return
        # If any of the values are set and are not pointing to files, save them
        # to a temporary directory
        try:
            with tempfile.TemporaryDirectory() as tls_dir:
                updated_files = {
                    key: val
                    for key, val in tls_opts.items()
                    if key not in non_file_opts and val
                }
                for fname, content in non_file_opts.items():
                    temp_fname = os.path.join(tls_dir, fname)
                    updated_files[fname] = temp_fname
                    with open(temp_fname, "w", encoding="utf-8") as handle:
                        handle.write(content)
                yield _TlsFiles(**updated_files)
        except OSError as err:
            log.error(
                "<RUN80977064E>",
                "Cannot create temporary TLS files. Either pass config as file paths or run with write permissions.",
                exc_info=True,
            )
            raise ValueError() from err


def main(blocking: bool = True):
    server = RuntimeHTTPServer()
    server._intercept_interrupt_signal()
    server.start(blocking)


if __name__ == "__main__":
    main()



# -*- coding: utf-8 -*-
# Generated by the protocol buffer compiler.  DO NOT EDIT!
# source: process.proto
"""Generated protocol buffer code."""
# Third Party
from google.protobuf import descriptor as _descriptor
from google.protobuf import descriptor_pool as _descriptor_pool
from google.protobuf import message as _message
from google.protobuf import reflection as _reflection
from google.protobuf import symbol_database as _symbol_database

# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()


DESCRIPTOR = _descriptor_pool.Default().AddSerializedFile(
    b'\n\rprocess.proto\x12\x0cprocessproto"\xfe\x02\n\x0eProcessRequest\x12\x12\n\ntrainingID\x18\x01 \x01(\t\x12\x18\n\x10\x63ustomTrainingID\x18\x06 \x01(\t\x12\x43\n\x0crequest_dict\x18\x02 \x03(\x0b\x32-.processproto.ProcessRequest.RequestDictEntry\x12\x1a\n\x12training_input_dir\x18\x03 \x01(\t\x12\x1b\n\x13training_output_dir\x18\x04 \x01(\t\x12\x16\n\x0e\x64\x61ta_input_dir\x18\x05 \x03(\t\x12\x41\n\x0b\x64\x61ta_inputs\x18\x07 \x03(\x0b\x32,.processproto.ProcessRequest.DataInputsEntry\x1a\x32\n\x10RequestDictEntry\x12\x0b\n\x03key\x18\x01 \x01(\t\x12\r\n\x05value\x18\x02 \x01(\t:\x02\x38\x01\x1a\x31\n\x0f\x44\x61taInputsEntry\x12\x0b\n\x03key\x18\x01 \x01(\t\x12\r\n\x05value\x18\x02 \x01(\t:\x02\x38\x01"\xbc\x01\n\x0fProcessResponse\x12\x12\n\ntrainingID\x18\x01 \x01(\t\x12\x18\n\x10\x63ustomTrainingID\x18\x03 \x01(\t\x12\x46\n\rresponse_dict\x18\x02 \x03(\x0b\x32/.processproto.ProcessResponse.ResponseDictEntry\x1a\x33\n\x11ResponseDictEntry\x12\x0b\n\x03key\x18\x01 \x01(\t\x12\r\n\x05value\x18\x02 \x01(\t:\x02\x38\x01\x32O\n\x07Process\x12\x44\n\x03Run\x12\x1c.processproto.ProcessRequest\x1a\x1d.processproto.ProcessResponse"\x00\x42\x64\n!com.ibm.watson.modeltrain.processZ?github.ibm.com/deep-learning-platform/processproto;processprotob\x06proto3'
)


_PROCESSREQUEST = DESCRIPTOR.message_types_by_name["ProcessRequest"]
_PROCESSREQUEST_REQUESTDICTENTRY = _PROCESSREQUEST.nested_types_by_name[
    "RequestDictEntry"
]
_PROCESSREQUEST_DATAINPUTSENTRY = _PROCESSREQUEST.nested_types_by_name[
    "DataInputsEntry"
]
_PROCESSRESPONSE = DESCRIPTOR.message_types_by_name["ProcessResponse"]
_PROCESSRESPONSE_RESPONSEDICTENTRY = _PROCESSRESPONSE.nested_types_by_name[
    "ResponseDictEntry"
]
ProcessRequest = _reflection.GeneratedProtocolMessageType(
    "ProcessRequest",
    (_message.Message,),
    {
        "RequestDictEntry": _reflection.GeneratedProtocolMessageType(
            "RequestDictEntry",
            (_message.Message,),
            {
                "DESCRIPTOR": _PROCESSREQUEST_REQUESTDICTENTRY,
                "__module__": "process_pb2"
                # @@protoc_insertion_point(class_scope:processproto.ProcessRequest.RequestDictEntry)
            },
        ),
        "DataInputsEntry": _reflection.GeneratedProtocolMessageType(
            "DataInputsEntry",
            (_message.Message,),
            {
                "DESCRIPTOR": _PROCESSREQUEST_DATAINPUTSENTRY,
                "__module__": "process_pb2"
                # @@protoc_insertion_point(class_scope:processproto.ProcessRequest.DataInputsEntry)
            },
        ),
        "DESCRIPTOR": _PROCESSREQUEST,
        "__module__": "process_pb2"
        # @@protoc_insertion_point(class_scope:processproto.ProcessRequest)
    },
)
_sym_db.RegisterMessage(ProcessRequest)
_sym_db.RegisterMessage(ProcessRequest.RequestDictEntry)
_sym_db.RegisterMessage(ProcessRequest.DataInputsEntry)

ProcessResponse = _reflection.GeneratedProtocolMessageType(
    "ProcessResponse",
    (_message.Message,),
    {
        "ResponseDictEntry": _reflection.GeneratedProtocolMessageType(
            "ResponseDictEntry",
            (_message.Message,),
            {
                "DESCRIPTOR": _PROCESSRESPONSE_RESPONSEDICTENTRY,
                "__module__": "process_pb2"
                # @@protoc_insertion_point(class_scope:processproto.ProcessResponse.ResponseDictEntry)
            },
        ),
        "DESCRIPTOR": _PROCESSRESPONSE,
        "__module__": "process_pb2"
        # @@protoc_insertion_point(class_scope:processproto.ProcessResponse)
    },
)
_sym_db.RegisterMessage(ProcessResponse)
_sym_db.RegisterMessage(ProcessResponse.ResponseDictEntry)

_PROCESS = DESCRIPTOR.services_by_name["Process"]
if _descriptor._USE_C_DESCRIPTORS == False:

    DESCRIPTOR._options = None
    DESCRIPTOR._serialized_options = b"\n!com.ibm.watson.modeltrain.processZ?github.ibm.com/deep-learning-platform/processproto;processproto"
    _PROCESSREQUEST_REQUESTDICTENTRY._options = None
    _PROCESSREQUEST_REQUESTDICTENTRY._serialized_options = b"8\001"
    _PROCESSREQUEST_DATAINPUTSENTRY._options = None
    _PROCESSREQUEST_DATAINPUTSENTRY._serialized_options = b"8\001"
    _PROCESSRESPONSE_RESPONSEDICTENTRY._options = None
    _PROCESSRESPONSE_RESPONSEDICTENTRY._serialized_options = b"8\001"
    _PROCESSREQUEST._serialized_start = 32
    _PROCESSREQUEST._serialized_end = 414
    _PROCESSREQUEST_REQUESTDICTENTRY._serialized_start = 313
    _PROCESSREQUEST_REQUESTDICTENTRY._serialized_end = 363
    _PROCESSREQUEST_DATAINPUTSENTRY._serialized_start = 365
    _PROCESSREQUEST_DATAINPUTSENTRY._serialized_end = 414
    _PROCESSRESPONSE._serialized_start = 417
    _PROCESSRESPONSE._serialized_end = 605
    _PROCESSRESPONSE_RESPONSEDICTENTRY._serialized_start = 554
    _PROCESSRESPONSE_RESPONSEDICTENTRY._serialized_end = 605
    _PROCESS._serialized_start = 607
    _PROCESS._serialized_end = 686
# @@protoc_insertion_point(module_scope)



# Copyright The Caikit Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""This module implements a LOCAL backend configuration
"""

# First Party
import alog

# Local
from ..exceptions import error_handler
from ..toolkit.wip_decorator import TempDisableWIP
from .backend_types import register_backend_type
from .base import BackendBase

log = alog.use_channel("LCLBKND")
error = error_handler.get(log)


class LocalBackend(BackendBase):
    backend_type = "LOCAL"

    def register_config(self, config) -> None:
        """Function to merge configs with existing configurations"""
        error(
            "<COR86557945E>",
            AssertionError(
                f"{self.backend_type} backend does not support this operation"
            ),
        )

    def start(self):
        """Start local backend. This is a no-op function"""
        self._started = True

    def stop(self):
        """Stop local backend. This is a no-op"""
        self._started = False


# Register local backend
with TempDisableWIP():
    register_backend_type(LocalBackend)



# Scheme Base (used for all schemes)
# General merged augmentor schemes to be leveraged by extensions of caikit core
# Local
from .always_selection_scheme import AlwaysSelectionScheme
from .base import SchemeBase
from .random_multi_selection_scheme import RandomMultiSelectionScheme
from .random_single_selection_scheme import RandomSingleSelectionScheme



# Copyright The Caikit Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""
Tests for the common factory implementation
"""

# Third Party
import pytest

# First Party
import aconfig

# Local
from caikit.core.toolkit import factory

## Helpers #####################################################################


class TypeOne(factory.FactoryConstructible):
    name = "one"

    def __init__(self, config, instance_name):
        self.config = config
        self.instance_name = instance_name


class TypeTwo(factory.FactoryConstructible):
    name = "two"

    def __init__(self, config, instance_name):
        self.config = config
        self.instance_name = instance_name


class TypeOtherOne(factory.FactoryConstructible):
    name = "one"

    def __init__(self, config, instance_name):
        self.config = config
        self.instance_name = instance_name


## Tests #######################################################################


def test_factory_happy_path():
    """Make sure that a factory works when used correctly"""
    fact = factory.Factory("Test")
    fact.register(TypeOne)
    fact.register(TypeTwo)

    inst_one = fact.construct({"type": "one"})
    assert isinstance(inst_one, TypeOne)
    assert isinstance(inst_one.config, aconfig.Config)

    inst_two = fact.construct({"type": "two", "config": {"foo": 1}})
    assert isinstance(inst_two, TypeTwo)
    assert isinstance(inst_two.config, aconfig.Config)
    assert inst_two.config.foo == 1


def test_factory_unregistered_error():
    """Make sure that asking to instantiate an unregistered type raises a
    ValueError
    """
    fact = factory.Factory("Test")
    with pytest.raises(ValueError):
        fact.construct({"type": "one"})


def test_factory_duplicate_registration():
    """Make sure that double registering a type is ok, but conflicting
    registration is not
    """
    fact = factory.Factory("Test")
    fact.register(TypeOne)
    fact.register(TypeOne)
    with pytest.raises(ValueError):
        fact.register(TypeOtherOne)


def test_factory_construct_with_instance_name():
    """Make sure that double registering a type is ok, but conflicting
    registration is not
    """
    fact = factory.Factory("Test")
    fact.register(TypeOne)

    inst_no_name = fact.construct({"type": "one"})
    assert isinstance(inst_no_name, TypeOne)
    assert inst_no_name.instance_name == TypeOne.name

    inst_name = "the-instance"
    inst_with_name = fact.construct({"type": "one"}, inst_name)
    assert isinstance(inst_with_name, TypeOne)
    assert inst_with_name.instance_name == inst_name



# Copyright The Caikit Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# Standard
try:
    # Standard
    from test.support.threading_helper import catch_threading_exception
except (NameError, ModuleNotFoundError):
    from tests.base import catch_threading_exception

# Standard
import threading
import time

# Third Party
import pytest

# Local
from caikit.core.toolkit.destroyable_thread import (
    DestroyableThread,
    ThreadDestroyedException,
)


def test_threads_can_be_interrupted():
    def infinite_wait():
        while True:
            time.sleep(0.001)

    thread = DestroyableThread(infinite_wait)
    thread.start()
    thread.destroy()
    assert thread.canceled
    thread.join(60)
    assert not thread.is_alive()


def test_threads_canceled_when_interrupt_fails():
    """This test exercises the case where the DestroyableThread fails to
    interrupt the thread with the internal exception on destroy(). The easiest
    way to simulate this is with a "long" `time.sleep` but that still doesn't
    fully guarantee that it will always be free of race conditions. In
    particular, the following two are still possible:

    1. The destroy() call lands in between waiting on the start event and
        beginning the time.sleep. This would cause the thread to die cleanly.
    2. The assertion that thread.is_alive() lands after the time.sleep has
        finished and the destroying exception has done its job to kill the
        thread.
    """
    start_event = threading.Event()
    end_event = threading.Event()

    def blocking_fn():
        start_event.set()
        time.sleep(0.01)
        end_event.wait()

    thread = DestroyableThread(blocking_fn)
    thread.start()
    start_event.wait()
    thread.destroy()
    assert thread.canceled
    # NOTE: We don't assert that thread.is_alive() here since it's potentially
    #   susceptible to the above mentioned race conditions. It _should_ always
    #   be true, though, based on reasonable timing.
    thread.join(60)
    assert not thread.is_alive()
    assert thread.canceled


def test_threads_can_catch_the_interrupts():
    started = threading.Event()
    caught = threading.Event()

    def test_catcher(started_event: threading.Event, caught_event: threading.Event):
        try:
            started_event.set()
            while True:
                time.sleep(0.001)
        except Exception as e:
            caught_event.set()
            raise e

    thread = DestroyableThread(test_catcher, started_event=started, caught_event=caught)

    thread.start()
    started.wait()

    thread.destroy()
    thread.join(60)

    assert not thread.is_alive()
    assert caught.is_set()


def test_threads_can_return_results():
    expected = "test-any-result"
    thread = DestroyableThread(lambda: expected)

    thread.start()
    thread.join()

    assert expected == thread.get_or_throw()


def test_threads_not_canceled_after_success():
    thread = DestroyableThread(lambda: None)
    thread.start()
    thread.join()
    assert not thread.canceled
    thread.destroy()
    assert not thread.canceled


def test_threads_can_throw():
    expected = ValueError("test-any-error")

    def thrower():
        raise expected

    thread = DestroyableThread(thrower)

    thread.start()
    thread.join()

    with pytest.raises(ValueError) as ctx:
        thread.get_or_throw()

    assert expected == ctx.value


def test_threads_will_not_execute_if_destroyed_before_starting():
    thread = DestroyableThread(lambda: time.sleep(1000))

    with catch_threading_exception() as cm:
        thread.destroy()
        thread.start()
        thread.join(1)

        assert not thread.is_alive()

        # Make sure the correct exception was raised
        assert cm.exc_type == ThreadDestroyedException


def test_event_is_set_on_completion():
    event = threading.Event()
    thread = DestroyableThread(lambda: None, work_done_event=event)

    assert not event.is_set()
    thread.start()
    thread.join()
    assert event.is_set()


def test_event_is_set_on_exception():
    event = threading.Event()

    def thrower():
        raise ValueError("test-any-exception")

    thread = DestroyableThread(thrower, work_done_event=event)

    assert not event.is_set()
    thread.start()
    thread.join()
    assert event.is_set()



# Copyright The Caikit Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

# Standard
from concurrent import futures
from typing import Optional, Union
import os

# Third Party
from grpc_health.v1 import health, health_pb2_grpc
from grpc_reflection.v1alpha import reflection
from py_grpc_prometheus.prometheus_server_interceptor import PromServerInterceptor
import grpc

# First Party
import aconfig
import alog

# Local
# Get the injectable servicer class definitions
from caikit import get_config
from caikit.runtime.interceptors.caikit_runtime_server_wrapper import (
    CaikitRuntimeServerWrapper,
)
from caikit.runtime.protobufs import (
    model_runtime_pb2,
    model_runtime_pb2_grpc,
    process_pb2_grpc,
)
from caikit.runtime.server_base import RuntimeServerBase
from caikit.runtime.service_factory import ServicePackage, ServicePackageFactory
from caikit.runtime.servicers.global_predict_servicer import GlobalPredictServicer
from caikit.runtime.servicers.global_train_servicer import GlobalTrainServicer
from caikit.runtime.servicers.model_runtime_servicer import ModelRuntimeServicerImpl
from caikit.runtime.servicers.model_train_servicer import ModelTrainServicerImpl
from caikit.runtime.servicers.training_management_servicer import (
    TrainingManagementServicerImpl,
)

# Have pylint ignore broad exception catching in this file so that we can log all
# unexpected errors using alog.
# pylint: disable=W0703
import caikit.core.data_model

log = alog.use_channel("SERVR-GRPC")
PROMETHEUS_METRICS_INTERCEPTOR = PromServerInterceptor()


class RuntimeGRPCServer(RuntimeServerBase):
    """An implementation of a gRPC server that serves caikit runtimes"""

    def __init__(
        self,
        tls_config_override: Optional[aconfig.Config] = None,
    ):
        super().__init__(get_config().runtime.grpc.port, tls_config_override)

        # Initialize basic server
        self.server = grpc.server(
            futures.ThreadPoolExecutor(
                max_workers=self.config.runtime.grpc.server_thread_pool_size
            ),
            interceptors=(PROMETHEUS_METRICS_INTERCEPTOR,),
            options=(self.config.runtime.grpc.options or {}).items(),
        )

        # Start metrics server
        RuntimeServerBase._start_metrics_server()

        # Start tracking service names for reflection
        service_names = [reflection.SERVICE_NAME]

        # Intercept an Inference Service
        self._global_predict_servicer = None
        if self.enable_inference:
            log.info("<RUN20247875I>", "Enabling gRPC inference service")
            self._global_predict_servicer = GlobalPredictServicer(
                self.inference_service
            )
            self.server = CaikitRuntimeServerWrapper(
                server=self.server,
                global_predict=self._global_predict_servicer.Predict,
                intercepted_svc_package=self.inference_service,
            )
            service_names.append(self.inference_service.descriptor.full_name)

            # Register inference service
            self.inference_service.registration_function(
                self.inference_service.service, self.server
            )

        # And intercept a training service, if we have one
        if self.enable_training and self.training_service:
            global_train_servicer = GlobalTrainServicer(self.training_service)
            self.server = CaikitRuntimeServerWrapper(
                server=self.server,
                global_predict=global_train_servicer.Train,
                intercepted_svc_package=self.training_service,
            )
            service_names.append(self.training_service.descriptor.full_name)

            # Register training service
            self.training_service.registration_function(
                self.training_service.service, self.server
            )

            # Add model train servicer to the gRPC server
            process_pb2_grpc.add_ProcessServicer_to_server(
                ModelTrainServicerImpl(self.training_service), self.server
            )

            # Add training management servicer to the gRPC server
            training_management_service: ServicePackage = (
                ServicePackageFactory.get_service_package(
                    ServicePackageFactory.ServiceType.TRAINING_MANAGEMENT,
                )
            )
            service_names.append(training_management_service.descriptor.full_name)

            training_management_service.registration_function(
                TrainingManagementServicerImpl(), self.server
            )

        # Add model runtime servicer to the gRPC server
        model_runtime_pb2_grpc.add_ModelRuntimeServicer_to_server(
            ModelRuntimeServicerImpl(), self.server
        )
        service_names.append(
            model_runtime_pb2.DESCRIPTOR.services_by_name["ModelRuntime"].full_name
        )

        # Add gRPC default health servicer.
        # We use the non-blocking implementation to avoid thread starvation.
        health_servicer = health.HealthServicer(
            experimental_non_blocking=True,
            experimental_thread_pool=futures.ThreadPoolExecutor(max_workers=1),
        )
        health_pb2_grpc.add_HealthServicer_to_server(health_servicer, self.server)

        # Finally enable service reflection after all services are added
        reflection.enable_server_reflection(service_names, self.server)

        # Listen on a unix socket as well for model mesh.
        if self.config.runtime.grpc.unix_socket_path and os.path.exists(
            self.config.runtime.grpc.unix_socket_path
        ):
            try:
                self.server.add_insecure_port(
                    f"unix://{self.config.runtime.grpc.unix_socket_path}"
                )
                log.info(
                    "<RUN10001011I>",
                    "Caikit Runtime is communicating through address: unix://%s",
                    self.config.runtime.grpc.unix_socket_path,
                )
            except RuntimeError:
                log.info(
                    "<RUN10001100I>",
                    "Binding failed for: unix://%s",
                    self.config.runtime.grpc.unix_socket_path,
                )

        if (
            self.tls_config
            and self.tls_config.server.key
            and self.tls_config.server.cert
        ):
            log.info("<RUN10001805I>", "Running with TLS")

            tls_server_pair = (
                bytes(self._load_secret(self.tls_config.server.key), "utf-8"),
                bytes(self._load_secret(self.tls_config.server.cert), "utf-8"),
            )
            if self.tls_config.client.cert:
                log.info("<RUN10001806I>", "Running with mutual TLS")
                # Client will verify the server using server cert and the server
                # will verify the client using client cert.
                server_credentials = grpc.ssl_server_credentials(
                    [tls_server_pair],
                    root_certificates=self._load_secret(self.tls_config.client.cert),
                    require_client_auth=True,
                )
            else:
                server_credentials = grpc.ssl_server_credentials([tls_server_pair])

            self.server.add_secure_port(f"[::]:{self.port}", server_credentials)
        else:
            log.info("<RUN10001807I>", "Running in insecure mode")
            self.server.add_insecure_port(f"[::]:{self.port}")

    def start(self, blocking: bool = True):
        """Boot the gRPC server. Can be non-blocking, or block until shutdown

        Args:
            blocking (boolean): Whether to block until shutdown
        """
        # Start the server. This is non-blocking, so we need to wait after
        self.server.start()

        log.info(
            "<RUN10001001I>",
            "Caikit Runtime is serving on port: %s with thread pool size: %s",
            self.port,
            self.config.runtime.grpc.server_thread_pool_size,
        )

        if blocking:
            self.server.wait_for_termination(None)

    def stop(self, grace_period_seconds: Union[float, int] = None):
        """Stop the server, with an optional grace period.

        Args:
            grace_period_seconds (Union[float, int]): Grace period for service shutdown.
                Defaults to application config
        """
        if grace_period_seconds is None:
            grace_period_seconds = (
                self.config.runtime.grpc.server_shutdown_grace_period_seconds
            )
        self.server.stop(grace_period_seconds)
        # Ensure we flush out any remaining billing metrics and stop metering
        if self.config.runtime.metering.enabled and self._global_predict_servicer:
            self._global_predict_servicer.stop_metering()
        # Shut down the model manager's model polling if enabled
        self._shut_down_model_manager()

    def render_protos(self, proto_out_dir: str) -> None:
        """Renders all the necessary protos for this service into a directory
        Args:
            proto_out_dir (str): Path to the directory to write proto files to
        """
        # First render out all the `@dataobject`s that should comprise the input / output
        # messages for these services
        caikit.core.data_model.render_dataobject_protos(proto_out_dir)

        # Then render each service
        if self.inference_service:
            self.inference_service.service.write_proto_file(proto_out_dir)
        if self.training_service:
            self.training_service.service.write_proto_file(proto_out_dir)

    def make_local_channel(self) -> grpc.Channel:
        """Return an insecure grpc channel over localhost for this server.
        Useful for unit testing or running local inference.
        """
        return grpc.insecure_channel(f"localhost:{self.port}")

    @staticmethod
    def _load_secret(secret: str) -> str:
        """If the secret points to a file, return the contents (plaintext reads).
        Else return the string"""
        if os.path.exists(secret):
            with open(secret, "r", encoding="utf-8") as secret_file:
                return secret_file.read()
        return secret

    # Context manager impl
    def __enter__(self):
        self.start(blocking=False)
        return self

    def __exit__(self, type_, value, traceback):
        self.stop(0)


def main(blocking: bool = True):
    server = RuntimeGRPCServer()
    server._intercept_interrupt_signal()
    server.start(blocking)


if __name__ == "__main__":
    main()



# Copyright The Caikit Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
Helpers for testing out data model functionality
"""

# Standard
from contextlib import contextmanager
from types import ModuleType
from typing import Dict, List, Optional, Union
import copy
import importlib
import os
import random
import re
import string
import sys
import tempfile

# Third Party
from google.protobuf import (
    descriptor,
    descriptor_pb2,
    descriptor_pool,
    struct_pb2,
    timestamp_pb2,
)

# First Party
import alog

# Local
from caikit.core.data_model.dataobject import _AUTO_GEN_PROTO_CLASSES
import caikit.core

log = alog.use_channel("TEST")


def add_fd_to_pool(
    fd: descriptor.FileDescriptor,
    dpool: descriptor_pool.DescriptorPool,
) -> List[descriptor_pb2.FileDescriptorProto]:
    """Helper to add a FileDescriptor to a new pool. This assumes the name is a
    unique identifier
    """
    added_fds = []
    try:
        dpool.FindFileByName(fd.name)
    except KeyError:
        log.debug2("Adding file %s to dpool %s", fd.name, dpool)
        for dep_fd in fd.dependencies:
            added_fds.extend(add_fd_to_pool(dep_fd, dpool))
        fd_proto = descriptor_pb2.FileDescriptorProto()
        fd.CopyToProto(fd_proto)
        dpool.Add(fd_proto)
        added_fds.append(fd_proto)
    return added_fds


@contextmanager
def temp_dpool(inherit_global: bool = False, skip_inherit: Optional[List[str]] = None):
    """Context to isolate the descriptor pool used in each test"""
    dpool = descriptor_pool.DescriptorPool()
    global_dpool = descriptor_pool._DEFAULT
    descriptor_pool._DEFAULT = dpool
    all_fd_protos = []
    all_fd_protos.extend(add_fd_to_pool(struct_pb2.DESCRIPTOR, dpool))
    all_fd_protos.extend(add_fd_to_pool(timestamp_pb2.DESCRIPTOR, dpool))

    # If inheriting from the current global, copy everything over
    if inherit_global:
        skip_inherit = skip_inherit or []
        for (
            dm_class
        ) in caikit.core.data_model.base._DataBaseMetaClass.class_registry.values():
            proto_class = dm_class.get_proto_class()
            fd_name = proto_class.DESCRIPTOR.file.name
            if any(re.match(skip_expr, fd_name) for skip_expr in skip_inherit):
                continue
            try:
                global_dpool.FindFileByName(fd_name)
                all_fd_protos.extend(add_fd_to_pool(proto_class.DESCRIPTOR.file, dpool))
            except KeyError:
                pass

    ##
    # HACK! Doing this _appears_ to solve the mysterious segfault cause by
    # using Struct inside a temporary descriptor pool. The inspiration for this
    # was:
    #
    # https://github.com/protocolbuffers/protobuf/issues/12047
    #
    # NOTE: This only works for protobuf 4.X (and as far as we know, it's not
    #     needed for 3.X)
    ##
    try:
        # Third Party
        from google.protobuf.message_factory import GetMessageClassesForFiles

        # Extra HACK! It seems that the below loop which _should_ do exactly
        # this somehow does not and by switching to it, the segfault reappears.
        msgs = GetMessageClassesForFiles(["google/protobuf/struct.proto"], dpool)
        _ = msgs["google.protobuf.Struct"]
        _ = msgs["google.protobuf.Value"]
        _ = msgs["google.protobuf.ListValue"]

        # for fd_proto in all_fd_protos:
        #     msgs = GetMessageClassesForFiles([fd_proto.name], dpool)
        #     for key in msgs:
        #         _ = msgs[key]

    # Nothing to do for protobuf 3.X
    except ImportError:
        pass
    yield dpool
    descriptor_pool._DEFAULT = global_dpool


def justify_script_string(script_str):
    """Helper to allow us to write 'scripts' as strings with space padding so
    that they don't look ugly!
    """
    min_indent = None
    lines = script_str.split("\n")
    for line in lines:
        if not line.strip():
            continue
        indent_len = len(line) - len(line.lstrip(" "))
        if min_indent is None or indent_len < min_indent:
            min_indent = indent_len
    if min_indent is not None:
        lines = [line[min_indent:] for line in lines]
    return "\n".join(lines)


def _make_data_model_module(
    module_dir: str, proto_defs: Union[str, List[str]]
) -> ModuleType:
    """Make a sample data model module!"""

    # Make the directory structure
    dm_dir = os.path.join(module_dir, "data_model")
    os.mkdir(dm_dir)

    # Set up the module hierarchy
    module_init = os.path.join(module_dir, "__init__.py")
    with open(module_init, "w", encoding="utf-8") as handle:
        handle.write(
            justify_script_string(
                """
                from . import data_model
                """
            )
        )
    dm_init = os.path.join(dm_dir, "__init__.py")
    if isinstance(proto_defs, str):
        proto_defs = [proto_defs]
    with open(dm_init, "w", encoding="utf-8") as handle:
        for proto_def in proto_defs:
            handle.write(proto_def)


@contextmanager
def temp_module():
    """Helper that will create an ephemeral module that can be imported and then
    clean it up afterwards
    """
    with tempfile.TemporaryDirectory() as workdir:
        # Make the randomized module name and directory
        mod_suffix = "".join(random.choice(string.ascii_lowercase) for i in range(10))
        mod_name = f"temp_mod_{mod_suffix}"
        mod_dir = os.path.join(workdir, mod_name)
        os.mkdir(mod_dir)

        # Add this the parent dir to the sys path so that it can be merged
        sys.path.append(workdir)

        # Yield the module directory and name so that the test can muck with it
        # before importing it
        yield mod_name, mod_dir

        # Remove the workdir from the sys path and delete the module from
        # sys.modules if it's there
        sys.path.pop()
        sys.modules.pop(mod_name, None)


@contextmanager
def temp_data_model(proto_defs: Union[str, List[str]]) -> ModuleType:
    """This contextmanager creates a temporary data model hierarchy with the
    given set of protobufs defs. It can be used to evaluate different
    combinations of message types without needing to manually write and compile
    protos for each test.
    """
    with temp_module() as (mod_name, module_dir):
        with temp_dpool():
            _make_data_model_module(module_dir, proto_defs)
            temp_mod = importlib.import_module(mod_name)
            yield temp_mod.data_model


def _random_package_suffix() -> str:
    some_list = list(
        "thisisabigoldstringofcharactersthatwillgetshuffledintoarandompackagesuffixthisisnotsuperb"
        "ombproofbutwhatever"
    )
    random.shuffle(some_list)
    return "".join(some_list)


def _get_proto_val_name(field_val) -> str:
    if isinstance(field_val, str):
        return field_val
    if isinstance(field_val, type):
        return field_val.__name__
    if isinstance(field_val, list):
        assert len(field_val) == 1
        return f"List[{_get_proto_val_name(field_val[0])}]"
    raise RuntimeError(f"Invalid field type specifier: {field_val}")


def make_proto_def(
    message_specs: Dict[str, dict],
    pkg_suffix: str = None,
    mock_compiled: bool = False,
) -> str:
    """Helper for writing a syntatically correct protobufs file"""
    if pkg_suffix is None:
        pkg_suffix = _random_package_suffix()
    package_name = f"{caikit.core.data_model.CAIKIT_DATA_MODEL}.{pkg_suffix}"
    out = justify_script_string(
        """
        from typing import Dict, List, Union
        """
    )
    if mock_compiled:
        out += justify_script_string(
            """
            from caikit.core.data_model import DataBase
            from py_to_proto import dataclass_to_proto, descriptor_to_message_class
            from dataclasses import dataclass
            """
        )
    else:
        out += "\nfrom caikit.core.data_model import DataObjectBase, dataobject\n"
    for message_name, message_spec in message_specs.items():
        type_annotations = ""
        for field_name, field_type in message_spec.items():
            type_annotations += f"    {field_name}: {_get_proto_val_name(field_type)}\n"

        # Manually create the proto class and use the precompiled proto style
        if mock_compiled:
            dataclass_name = f"_{message_name}"
            proto_name = f"{dataclass_name}_proto"
            msg_str = f"\n@dataclass\nclass {dataclass_name}:\n"
            msg_str += type_annotations
            msg_str += justify_script_string(
                f"""

                {proto_name} = descriptor_to_message_class(
                    dataclass_to_proto(
                        dataclass_={dataclass_name},
                        package="{package_name}",
                        name="{message_name}",
                    )
                )

                class {message_name}(DataBase):
                    _proto_class = {proto_name}
                """
            )

        # Add as a dataobject
        else:
            msg_str = f'\n@dataobject("{package_name}")\nclass {message_name}(DataObjectBase):\n'
            msg_str += type_annotations

        out += msg_str
    return out


@contextmanager
def reset_global_protobuf_registry():
    """Reset the global registry of generated protos"""
    prev_auto_gen_proto_classes = copy.copy(_AUTO_GEN_PROTO_CLASSES)
    prev_class_registry = copy.copy(
        caikit.core.data_model.base._DataBaseMetaClass.class_registry
    )
    _AUTO_GEN_PROTO_CLASSES.clear()
    yield
    _AUTO_GEN_PROTO_CLASSES.extend(prev_auto_gen_proto_classes)
    caikit.core.data_model.base._DataBaseMetaClass.class_registry.clear()
    caikit.core.data_model.base._DataBaseMetaClass.class_registry.update(
        prev_class_registry
    )



# Copyright The Caikit Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
This test suite ensures that any model metadata (things inside a config.yml file) is persisted when
a module is loaded and then re-saved
"""

# Standard
from typing import Any, Dict, Set
import os
import tempfile

# Third Party
import pytest

# Local
from caikit.core import toolkit

# pylint: disable=import-error
from sample_lib.data_model import SampleTask

# pylint: disable=import-error
from sample_lib.modules.sample_task import SampleModule

# Unit Test Infrastructure
from tests.base import TestCaseBase
import caikit.core

# scratch:
# - might want to support like a `saved` date?
# - want a unique id
# - do we want to easily support editing metadata in the case that somebody:
#   - loads a model
#   - changes the python model object
#   - re-saves the model?
# - What about the `version` field? It's unused now but if it does change should we allow it to
# re-save with a new version?


def _load_model_metadata(model_path: str) -> Dict[str, Any]:
    return toolkit.load_yaml(os.path.join(model_path, "config.yml"))


def _check_dicts_equal(
    this: Dict[str, Any], that: Dict[str, Any], fields_to_not_check: Set[str]
):
    these_fields = set(this.keys()) - fields_to_not_check
    those_fields = set(that.keys()) - fields_to_not_check
    assert these_fields == those_fields

    for field in these_fields:
        assert this.get(field) == that.get(field)


@pytest.fixture
def fixtures_dir():
    return TestCaseBase.fixtures_dir


@pytest.fixture
# pylint: disable=redefined-outer-name
def sample_model_path(fixtures_dir):
    return os.path.join(fixtures_dir, "sample_module")


# pylint: disable=redefined-outer-name
def test_module_metadata_is_persisted(sample_model_path):
    """Make sure that if you load and then re-save any model, the metadata in the config.yml is
    persisted."""
    # Get the module metadata:
    initial_metadata = _load_model_metadata(sample_model_path)
    model = caikit.core.load(sample_model_path)

    with tempfile.TemporaryDirectory() as tempdir:
        model.save(tempdir)
        resaved_metadata = _load_model_metadata(tempdir)

    fields_to_not_check = {"saved", "sample_lib_version", "train"}
    _check_dicts_equal(initial_metadata, resaved_metadata, fields_to_not_check)

    # Fun little extra check, the `train` field _should_ be modified by the code
    assert "optim_method" not in resaved_metadata["train"]


# pylint: disable=redefined-outer-name
def test_loaded_modules_have_metadata(sample_model_path):
    """Make sure a model has metadata after being loaded"""
    expected_metadata = _load_model_metadata(sample_model_path)
    model_loaded_with_core = caikit.core.load(sample_model_path)
    model_directly_loaded = SampleModule.load(sample_model_path)

    # TODO: figure out "module_id" and "model_path" as well...

    _check_dicts_equal(
        expected_metadata,
        model_loaded_with_core.metadata,
        {"module_id", "model_path"},
    )
    _check_dicts_equal(
        expected_metadata,
        model_directly_loaded.metadata,
        {"module_id", "model_path"},
    )


def test_module_has_saved_field():
    """Make sure that if you load a model and then save it multiple times,
    the "saved" field should track each timestamp when you save, and should be different
    """
    with tempfile.TemporaryDirectory() as tempdir:
        model1 = SampleModule()
        path1 = os.path.join(tempdir, "test1")
        model1.save(path1)
        resaved_metadata1 = _load_model_metadata(path1)

        model2 = caikit.core.load(path1)
        path2 = os.path.join(tempdir, "test1")
        model2.save(path2)
        resaved_metadata2 = _load_model_metadata(path2)

    assert "saved" in resaved_metadata1
    assert "saved" in resaved_metadata2
    assert resaved_metadata1["saved"] != resaved_metadata2["saved"]


def test_module_has_tracking_id_field():
    with tempfile.TemporaryDirectory() as tempdir:
        model1 = SampleModule()
        path1 = os.path.join(tempdir, "test1")
        model1.save(path1)
        resaved_metadata1 = _load_model_metadata(path1)

        model2 = caikit.core.load(path1)
        path2 = os.path.join(tempdir, "test1")
        model2.save(path2)
        resaved_metadata2 = _load_model_metadata(path2)

    assert "tracking_id" in resaved_metadata1
    assert "tracking_id" in resaved_metadata2
    assert resaved_metadata1["tracking_id"] == resaved_metadata2["tracking_id"]


# pylint: disable=redefined-outer-name
def test_load_can_be_called_directly_with_non_standard_kwargs(sample_model_path):
    initial_metadata = _load_model_metadata(sample_model_path)
    # note that
    # - no positional arguments given
    # - path is `model_path` not `module_path`
    model = SampleModule.load(foo="bar", test_kw="arg", model_path=sample_model_path)

    assert len(model.metadata) > 0
    _check_dicts_equal(initial_metadata, model.metadata, {"module_id", "model_path"})

    # Write a class that doesn't have a `xxx_path` arg for load
    @caikit.core.module(
        "00110203-0809-beef-baad-0a0b0c0d0e0f", "FunkyModule", "0.0.1", SampleTask
    )
    class _FunkyModel(SampleModule):
        @classmethod
        def load(cls, some_really_odd_param_name):
            return super().load(some_really_odd_param_name)

    # check this doesn't raise
    # (it won't extract metadata though...)
    _FunkyModel.load(some_really_odd_param_name=sample_model_path)


# pylint: disable=redefined-outer-name
def test_parent_class_loads_work(sample_model_path):
    """This test ensures that our metadata injector works on modules that inherit from other
    classes"""
    model = SampleModule.load(sample_model_path)

    assert isinstance(model, SampleModule)



# Copyright The Caikit Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""
Tests for the caikit HTTP server
"""
# Standard
from contextlib import contextmanager
from dataclasses import dataclass
from typing import Dict
import json
import os
import signal
import tempfile

# Third Party
from fastapi.testclient import TestClient
import numpy as np
import pytest
import requests
import tls_test_tools

# Local
from caikit.core import MODEL_MANAGER, DataObjectBase, dataobject
from caikit.runtime import http_server
from tests.conftest import temp_config
from tests.runtime.conftest import (
    ModuleSubproc,
    register_trained_model,
    runtime_http_test_server,
)

## Helpers #####################################################################


def save_key_cert_pair(prefix, workdir, key=None, cert=None):
    crtfile, keyfile = None, None
    if key is not None:
        keyfile = os.path.join(workdir, f"{prefix}.key")
        with open(keyfile, "w") as handle:
            handle.write(key)
    if cert is not None:
        crtfile = os.path.join(workdir, f"{prefix}.crt")
        with open(crtfile, "w") as handle:
            handle.write(cert)
    return crtfile, keyfile


@dataobject
class KeyPair(DataObjectBase):
    cert: str
    key: str


@dataobject
class TLSConfig(DataObjectBase):
    server: KeyPair
    client: KeyPair


@contextmanager
def generate_tls_configs(
    port: int,
    tls: bool = False,
    mtls: bool = False,
    inline: bool = False,
    **http_config_overrides,
) -> Dict[str, Dict]:
    """Helper to generate tls configs"""
    with tempfile.TemporaryDirectory() as workdir:
        config_overrides = {}
        client_keyfile, client_certfile, ca_certfile = None, None, None
        ca_cert, server_cert, server_key = None, None, None
        if mtls or tls:
            ca_key = tls_test_tools.generate_key()[0]
            ca_cert = tls_test_tools.generate_ca_cert(ca_key)
            ca_certfile, _ = save_key_cert_pair("ca", workdir, cert=ca_cert)
            server_key, server_cert = tls_test_tools.generate_derived_key_cert_pair(
                ca_key=ca_key
            )
            server_certfile, server_keyfile = save_key_cert_pair(
                "server", workdir, server_key, server_cert
            )

            if inline:
                tls_config = TLSConfig(
                    server=KeyPair(cert=server_cert, key=server_key),
                    client=KeyPair(cert=ca_cert if mtls else "", key=""),
                )
            else:
                tls_config = TLSConfig(
                    server=KeyPair(cert=server_certfile, key=server_keyfile),
                    client=KeyPair(cert=ca_certfile if mtls else "", key=""),
                )
            # need to save this ca_certfile in config_overrides so the tls tests below can access it from client side
            config_overrides["use_in_test"] = {"ca_cert": ca_certfile}

            # also saving a bad ca_certfile for a failure test case
            bad_ca_file = os.path.join(workdir, "bad_ca_cert.crt")
            with open(bad_ca_file, "w") as handle:
                bad_cert = (
                    "-----BEGIN CERTIFICATE-----\nfoobar\n-----END CERTIFICATE-----"
                )
                handle.write(bad_cert)
            config_overrides["use_in_test"]["bad_ca_cert"] = bad_ca_file

            if mtls:
                client_certfile, client_keyfile = save_key_cert_pair(
                    "client",
                    workdir,
                    *tls_test_tools.generate_derived_key_cert_pair(ca_key=ca_key),
                )
                # need to save the client cert and key in config_overrides so the mtls test below can access it
                config_overrides["use_in_test"]["client_cert"] = client_certfile
                config_overrides["use_in_test"]["client_key"] = client_keyfile

            config_overrides["runtime"] = {"tls": tls_config.to_dict()}
        config_overrides.setdefault("runtime", {})["http"] = {
            "server_shutdown_grace_period_seconds": 0.01,  # this is so the server is killed after 0.1 if no test is running
            "port": port,
            **http_config_overrides,
        }

        with temp_config(config_overrides, "merge"):
            yield config_overrides


## Insecure and TLS Tests #######################################################################


def test_insecure_server(runtime_http_server, open_port):
    with generate_tls_configs(open_port):
        # start a non-blocking http server
        resp = requests.get(f"http://localhost:{runtime_http_server.port}/docs")
        resp.raise_for_status()


def test_basic_tls_server(open_port):
    with generate_tls_configs(
        open_port, tls=True, mtls=False, http_config_overrides={}
    ) as config_overrides:
        with runtime_http_test_server(
            open_port,
            tls_config_override=config_overrides,
        ) as http_server_with_tls:
            # start a non-blocking http server with basic tls
            resp = requests.get(
                f"https://localhost:{http_server_with_tls.port}/docs",
                verify=config_overrides["use_in_test"]["ca_cert"],
            )
            resp.raise_for_status()


def test_basic_tls_server_with_wrong_cert(open_port):
    with generate_tls_configs(
        open_port, tls=True, mtls=False, http_config_overrides={}
    ) as config_overrides:
        with runtime_http_test_server(
            open_port,
            tls_config_override=config_overrides,
        ) as http_server_with_tls:
            # start a non-blocking http server with basic tls
            with pytest.raises(requests.exceptions.SSLError):
                requests.get(
                    f"https://localhost:{http_server_with_tls.port}/docs",
                    verify=config_overrides["use_in_test"]["bad_ca_cert"],
                )


def test_mutual_tls_server(open_port):
    with generate_tls_configs(
        open_port, tls=True, mtls=True, http_config_overrides={}
    ) as config_overrides:
        with runtime_http_test_server(
            open_port,
            tls_config_override=config_overrides,
        ) as http_server_with_mtls:
            # start a non-blocking http server with mutual tls
            resp = requests.get(
                f"https://localhost:{http_server_with_mtls.port}/docs",
                verify=config_overrides["use_in_test"]["ca_cert"],
                cert=(
                    config_overrides["use_in_test"]["client_cert"],
                    config_overrides["use_in_test"]["client_key"],
                ),
            )
            resp.raise_for_status()


def test_mutual_tls_server_inline(open_port):
    """Test that mutual TLS works when the TLS content is passed by value rather
    than with files
    """
    with generate_tls_configs(
        open_port, tls=True, mtls=True, inline=True, http_config_overrides={}
    ) as config_overrides:
        with runtime_http_test_server(
            open_port,
            tls_config_override=config_overrides,
        ) as http_server_with_mtls:
            # start a non-blocking http server with mutual tls
            resp = requests.get(
                f"https://localhost:{http_server_with_mtls.port}/docs",
                verify=config_overrides["use_in_test"]["ca_cert"],
                cert=(
                    config_overrides["use_in_test"]["client_cert"],
                    config_overrides["use_in_test"]["client_key"],
                ),
            )
            resp.raise_for_status()


def test_mutual_tls_server_with_wrong_cert(open_port):
    with generate_tls_configs(
        open_port, tls=True, mtls=True, http_config_overrides={}
    ) as config_overrides:
        with runtime_http_test_server(
            open_port,
            tls_config_override=config_overrides,
        ) as http_server_with_mtls:
            # start a non-blocking http server with mutual tls
            with pytest.raises(requests.exceptions.SSLError):
                requests.get(
                    f"https://localhost:{http_server_with_mtls.port}/docs",
                    verify=config_overrides["use_in_test"]["ca_cert"],
                    cert=(
                        config_overrides["use_in_test"]["client_key"],
                        config_overrides["use_in_test"]["client_cert"],
                    ),  # flip the order of key and cert, this will result in SSLError
                )


@pytest.mark.parametrize(
    "enabled_services",
    [(True, False), (False, True), (False, False)],
)
def test_services_disabled(open_port, enabled_services):
    enable_inference, enable_training = enabled_services
    with temp_config(
        {
            "runtime": {
                "service_generation": {
                    "enable_inference": enable_inference,
                    "enable_training": enable_training,
                }
            },
        },
        "merge",
    ):
        with runtime_http_test_server(open_port) as server:
            # start a non-blocking http server with basic tls
            resp = requests.get(
                f"http://localhost:{open_port}{http_server.HEALTH_ENDPOINT}",
            )
            resp.raise_for_status()
            assert server.enable_inference == enable_inference
            assert (server.global_predict_servicer and enable_inference) or (
                server.global_predict_servicer is None and not enable_inference
            )
            assert server.enable_training == enable_training
            # TODO: Update once training enabled
            # assert (server.global_train_servicer and enable_training) or (
            #     server.global_train_servicer is None and not enable_training
            # )


## Inference Tests #######################################################################


def test_docs(runtime_http_server):
    """Simple check that pinging /docs returns 200"""
    with TestClient(runtime_http_server.app) as client:
        response = client.get("/docs")
        assert response.status_code == 200


def test_docs_using_running_http_server(runtime_http_server):
    """Simple check that pinging /docs returns 200
    but pints the actual running server"""
    response = requests.get(f"http://localhost:{runtime_http_server.port}/docs")
    assert response.status_code == 200


def test_inference_sample_task(sample_task_model_id, runtime_http_server):
    """Simple check that we can ping a model"""
    with TestClient(runtime_http_server.app) as client:
        json_input = {"inputs": {"name": "world"}}
        response = client.post(
            f"/api/v1/{sample_task_model_id}/task/sample",
            json=json_input,
        )
        json_response = json.loads(response.content.decode(response.default_encoding))
        assert response.status_code == 200, json_response
        assert json_response["greeting"] == "Hello world"


def test_inference_sample_task_optional_field(
    sample_task_model_id, runtime_http_server
):
    """Simple check for optional fields"""
    with TestClient(runtime_http_server.app) as client:
        json_input = {
            "inputs": {"name": "world"},
            "parameters": {"throw": True},
        }
        response = client.post(
            f"/api/v1/{sample_task_model_id}/task/sample",
            json=json_input,
        )
        # this is 500 because we explicitly pass in `throw` as True, which
        # raises an internal error in the module
        assert response.status_code == 500


def test_inference_other_task(other_task_model_id, runtime_http_server):
    """Simple check that we can ping a model"""
    with TestClient(runtime_http_server.app) as client:
        json_input = {"inputs": {"name": "world"}}
        response = client.post(
            f"/api/v1/{other_task_model_id}/task/other",
            json=json_input,
        )
        json_response = json.loads(response.content.decode(response.default_encoding))
        assert response.status_code == 200, json_response
        assert json_response["farewell"] == "goodbye: world 42 times"


def test_json_file_task(file_task_model_id, runtime_http_server):
    """Simple check that we can ping a model"""
    with TestClient(runtime_http_server.app) as client:
        # cGRmZGF0Yf//AA== is b"pdfdata\xff\xff\x00" base64 encoded
        json_input = {"inputs": {"filename": "example.pdf", "data": "cGRmZGF0Yf//AA=="}}

        response = client.post(
            f"/api/v1/{file_task_model_id}/task/file",
            json=json_input,
        )
        json_response = json.loads(response.content.decode(response.default_encoding))
        assert response.status_code == 200, json_response
        assert json_response["filename"] == "processed_example.pdf"
        # Ym91bmRpbmd8cGRmZGF0Yf//AHxib3g= is b"bounding|pdfdata\xff\xff\x00|box" base64 encoded
        assert json_response["data"] == "Ym91bmRpbmd8cGRmZGF0Yf//AHxib3g="


def test_inference_streaming_sample_module(sample_task_model_id, runtime_http_server):
    """Simple check for testing a happy path unary-stream case"""
    with TestClient(runtime_http_server.app) as client:
        json_input = {"inputs": {"name": "world"}}
        stream = client.post(
            f"/api/v1/{sample_task_model_id}/task/server-streaming-sample",
            json=json_input,
        )
        assert stream.status_code == 200
        stream_content = stream.content.decode(stream.default_encoding)
        stream_responses = json.loads(
            "[{}]".format(
                stream_content.replace("data: ", "")
                .replace("\r\n", "")
                .replace("}{", "}, {")
            )
        )
        assert len(stream_responses) == 10
        assert all(
            resp.get("greeting") == "Hello world stream" for resp in stream_responses
        )


def test_model_not_found(runtime_http_server):
    """Simple check that we can ping a model"""
    with TestClient(runtime_http_server.app) as client:
        response = client.post(
            f"/api/v1/this_is_not_a_model/task/sample",
            json={"inputs": {"name": "world"}},
        )
        assert response.status_code == 404


def test_inference_sample_task_incorrect_input(
    sample_task_model_id, runtime_http_server
):
    """Test that with an incorrect input, the test doesn't throw but
    instead returns None"""
    with TestClient(runtime_http_server.app) as client:
        json_input = {
            "inputs": {"blah": "world"},
        }
        response = client.post(
            f"/api/v1/{sample_task_model_id}/task/sample",
            json=json_input,
        )
        assert response.status_code == 422, response.content.decode(
            response.default_encoding
        )


@pytest.mark.skip("Skipping since we're not tacking forward compatibility atm")
def test_inference_sample_task_forward_compatibility(
    sample_task_model_id, runtime_http_server
):
    """Test that clients can send in params that don't exist on server
    without any error"""
    with TestClient(runtime_http_server.app) as client:
        json_input = {
            "inputs": {"name": "world", "blah": "blah"},
        }
        response = client.post(
            f"/api/v1/{sample_task_model_id}/task/sample",
            json=json_input,
        )
        json_response = json.loads(response.content.decode(response.default_encoding))
        assert response.status_code == 200, json_response
        assert json_response["greeting"] == "Hello world"


def test_health_check_ok(runtime_http_server):
    """Make sure the health check returns OK"""
    with TestClient(runtime_http_server.app) as client:
        response = client.get(http_server.HEALTH_ENDPOINT)
        assert response.status_code == 200
        assert response.text == "OK"


def test_http_server_shutdown_with_model_poll(open_port):
    """Test that a SIGINT successfully shuts down the running server"""
    with tempfile.TemporaryDirectory() as workdir:
        server_proc = ModuleSubproc(
            "caikit.runtime.http_server",
            RUNTIME_HTTP_PORT=str(open_port),
            RUNTIME_LOCAL_MODELS_DIR=workdir,
            RUNTIME_LAZY_LOAD_LOCAL_MODELS="true",
            RUNTIME_LAZY_LOAD_POLL_PERIOD_SECONDS="0.1",
        )
        with server_proc as proc:
            # Wait for the server to be up
            while True:
                try:
                    resp = requests.get(
                        f"http://localhost:{open_port}{http_server.HEALTH_ENDPOINT}",
                        timeout=0.1,
                    )
                    resp.raise_for_status()
                    break
                except (
                    requests.HTTPError,
                    requests.ConnectionError,
                    requests.ConnectTimeout,
                ):
                    pass

            # Signal the server to shut down
            proc.send_signal(signal.SIGINT)

        # Make sure the process was not killed
        assert not server_proc.killed


## Train Tests #######################################################################


def test_train_sample_task(runtime_http_server):
    model_name = "sample_task_train"
    with TestClient(runtime_http_server.app) as client:
        json_input = {
            "model_name": model_name,
            "parameters": {
                "training_data": {"data_stream": {"data": [{"number": 1}]}},
                "batch_size": 42,
            },
        }
        training_response = client.post(
            f"/api/v1/SampleTaskSampleModuleTrain",
            json=json_input,
        )

        # assert training response
        training_json_response = json.loads(
            training_response.content.decode(training_response.default_encoding)
        )
        assert training_response.status_code == 200, training_json_response
        assert (training_id := training_json_response["training_id"])
        assert training_json_response["model_name"] == model_name

        # assert trained model
        result = MODEL_MANAGER.get_model_future(training_id).load()
        assert result.batch_size == 42
        assert (
            result.MODULE_CLASS
            == "sample_lib.modules.sample_task.sample_implementation.SampleModule"
        )

        # register the newly trained model for inferencing
        register_trained_model(
            runtime_http_server.global_predict_servicer,
            model_name,
            training_id,
        )

        # test inferencing on new model
        json_input_inference = {"inputs": {"name": "world"}}
        response = client.post(
            f"/api/v1/{model_name}/task/sample",
            json=json_input_inference,
        )
        json_response = json.loads(response.content.decode(response.default_encoding))
        assert response.status_code == 200, json_response
        assert json_response["greeting"] == "Hello world"


def test_train_sample_task_throws_s3_value_error(runtime_http_server):
    """test that if we provide s3 path, it throws an error"""
    model_name = "sample_task_train"
    with TestClient(runtime_http_server.app) as client:
        json_input = {
            "model_name": model_name,
            "output_path": {"path": "non-existent path_to_s3"},
            "parameters": {
                "training_data": {"data_stream": {"data": [{"number": 1}]}},
                "batch_size": 42,
            },
        }
        training_response = client.post(
            f"/api/v1/SampleTaskSampleModuleTrain",
            json=json_input,
        )
        assert (
            "S3 output path not supported by this runtime"
            in training_response.content.decode(training_response.default_encoding)
        )
        assert training_response.status_code == 500, training_response.content.decode(
            training_response.default_encoding
        )


def test_train_primitive_task(runtime_http_server):
    model_name = "primitive_task_train"
    with TestClient(runtime_http_server.app) as client:
        json_input = {
            "model_name": model_name,
            "parameters": {
                "sample_input": {"name": "test"},
                "simple_list": ["hello", "world"],
                "union_list": ["hello", "world"],
                "union_list2": ["hello", "world"],
                "union_list3": ["hello", "world"],
                "union_list4": 1,
                "training_params_json_dict_list": [{"foo": {"bar": [1, 2, 3]}}],
                "training_params_json_dict": {"foo": {"bar": [1, 2, 3]}},
                "training_params_dict": {"layer_sizes": 100, "window_scaling": 200},
                "training_params_dict_int": {1: 0.1, 2: 0.01},
            },
        }

        training_response = client.post(
            f"/api/v1/SampleTaskSamplePrimitiveModuleTrain",
            json=json_input,
        )
        # assert training response
        training_json_response = json.loads(
            training_response.content.decode(training_response.default_encoding)
        )
        assert training_response.status_code == 200, training_json_response
        assert (training_id := training_json_response["training_id"])
        assert training_json_response["model_name"] == model_name

        # assert trained model
        result = MODEL_MANAGER.get_model_future(training_id).load()
        assert result.training_params_dict == {
            "layer_sizes": 100,
            "window_scaling": 200,
        }
        assert result.training_params_json_dict == {"foo": {"bar": [1, 2, 3]}}
        assert (
            result.MODULE_CLASS
            == "sample_lib.modules.sample_task.primitive_party_implementation.SamplePrimitiveModule"
        )

        # register the newly trained model for inferencing
        register_trained_model(
            runtime_http_server.global_predict_servicer,
            model_name,
            training_id,
        )

        # test inferencing on new model
        json_input_inference = {"inputs": {"name": "world"}}
        response = client.post(
            f"/api/v1/{model_name}/task/sample",
            json=json_input_inference,
        )
        json_response = json.loads(response.content.decode(response.default_encoding))
        assert response.status_code == 200, json_response
        assert json_response["greeting"] == "hello: primitives! [1, 2, 3] 100"


def test_train_other_task(runtime_http_server):
    model_name = "other_task_train"
    with TestClient(runtime_http_server.app) as client:
        json_input = {
            "model_name": model_name,
            "parameters": {
                "training_data": {"data_stream": {"data": [1, 2]}},
                "sample_input": {"name": "test"},
            },
        }

        training_response = client.post(
            f"/api/v1/OtherTaskOtherModuleTrain",
            json=json_input,
        )
        # assert training response
        training_json_response = json.loads(
            training_response.content.decode(training_response.default_encoding)
        )
        assert training_response.status_code == 200, training_json_response
        assert (training_id := training_json_response["training_id"])
        assert training_json_response["model_name"] == model_name

        # assert trained model
        result = MODEL_MANAGER.get_model_future(training_id).load()
        assert result.batch_size == 64
        assert (
            result.MODULE_CLASS
            == "sample_lib.modules.other_task.other_implementation.OtherModule"
        )

        # register the newly trained model for inferencing
        register_trained_model(
            runtime_http_server.global_predict_servicer,
            model_name,
            training_id,
        )

        # test inferencing on new model
        json_input_inference = {"inputs": {"name": "world"}}
        response = client.post(
            f"/api/v1/{model_name}/task/other",
            json=json_input_inference,
        )
        json_response = json.loads(response.content.decode(response.default_encoding))
        assert response.status_code == 200, json_response
        assert json_response["farewell"] == "goodbye: world 64 times"
